import copy as _copy
import math as _math
import os as _os

import cv2 as _cv2
import numpy as _np
from easytorch.utils.logger import *
from easytorch.data.multiproc import multiRun
from PIL import Image as _IMG
import json as _json
import traceback as _tb

"""
##################################################################################################
Very useful image related utilities
##################################################################################################
"""


def _same_file(x):
    return x


class Image:
    def __init__(self, dtype=_np.uint8):
        self.dir = None
        self.file = None
        self.array = None
        self.mask = None
        self.ground_truth = None
        self.extras = {}
        self.dtype = dtype

    def load(self, dir, file):
        try:
            self.dir = dir
            self.file = file
            self.array = _np.array(_IMG.open(self.path), dtype=self.dtype)
        except Exception as e:
            error('Fail to load file: ' + self.file + ': ' + str(e))

    def load_mask(self, mask_dir=None, fget_mask=_same_file):
        if fget_mask is None:
            fget_mask = _same_file
        try:
            mask_file = fget_mask(self.file)
            self.mask = _np.array(_IMG.open(_os.path.join(mask_dir, mask_file)), dtype=self.dtype)
        except Exception as e:
            error('Fail to load mask: ' + str(e))

    def load_ground_truth(self, gt_dir=None, fget_ground_truth=_same_file):
        if fget_ground_truth is None:
            fget_ground_truth = _same_file
        try:
            gt_file = fget_ground_truth(self.file)
            self.ground_truth = _np.array(_IMG.open(_os.path.join(gt_dir, gt_file)), dtype=self.dtype)
        except Exception as e:
            error('Fail to load ground truth: ' + str(e))

    def get_array(self, dir='', getter=_same_file, file=None):
        if getter is None:
            getter = _same_file
        if not file:
            file = self.file
        arr = _np.array(_IMG.open(_os.path.join(dir, getter(file))), dtype=self.dtype)
        return arr

    def apply_mask(self):
        if self.mask is not None:
            self.array[self.mask == 0] = 0

    def apply_clahe(self, clip_limit=2.0, tile_shape=(8, 8)):
        enhancer = _cv2.createCLAHE(clipLimit=clip_limit, tileGridSize=tile_shape)
        if len(self.array.shape) == 2:
            self.array = enhancer.apply(self.array)
        elif len(self.array.shape) == 3:
            self.array[:, :, 0] = enhancer.apply(self.array[:, :, 0])
            self.array[:, :, 1] = enhancer.apply(self.array[:, :, 1])
            self.array[:, :, 2] = enhancer.apply(self.array[:, :, 2])
        else:
            error('More than three channels')

    def copy(self):
        copy_obj = Image()
        copy_obj.file = _copy.copy(self.file)
        copy_obj.array = _copy.copy(self.array)
        copy_obj.mask = _copy.copy(self.mask)
        copy_obj.ground_truth = _copy.copy(self.ground_truth)
        copy_obj.dtype = _copy.copy(self.dtype)
        copy_obj.extras = _copy.deepcopy(self.extras)
        return copy_obj

    @property
    def path(self):
        return _os.path.join(self.dir, self.file)


def get_rgb_scores(arr_2d=None, truth=None):
    """
    Returns a rgb image of pixelwise separation between ground truth and arr_2d
    (predicted image) with different color codes
    Easy when needed to inspect segmentation result against ground truth.
    :param arr_2d:
    :param truth:
    :return:
    """
    arr_rgb = _np.zeros([arr_2d.shape[0], arr_2d.shape[1], 3], dtype=_np.uint8)
    x = arr_2d.copy()
    y = truth.copy()
    x[x == 255] = 1
    y[y == 255] = 1
    xy = x + (y * 2)
    arr_rgb[xy == 3] = [255, 255, 255]
    arr_rgb[xy == 1] = [0, 255, 0]
    arr_rgb[xy == 2] = [255, 0, 0]
    arr_rgb[xy == 0] = [0, 0, 0]
    return arr_rgb


def get_praf1(arr_2d=None, truth=None):
    """
    Returns precision, recall, f1 and accuracy score between two binary arrays upto five precision.
    :param arr_2d:
    :param truth:
    :return:
    """
    x = arr_2d.copy()
    y = truth.copy()
    x[x == 255] = 1
    y[y == 255] = 1
    xy = x + (y * 2)
    tp = xy[xy == 3].shape[0]
    fp = xy[xy == 1].shape[0]
    tn = xy[xy == 0].shape[0]
    fn = xy[xy == 2].shape[0]
    try:
        p = tp / (tp + fp)
    except ZeroDivisionError:
        p = 0

    try:
        r = tp / (tp + fn)
    except ZeroDivisionError:
        r = 0

    try:
        a = (tp + tn) / (tp + fp + fn + tn)
    except ZeroDivisionError:
        a = 0

    try:
        f1 = 2 * p * r / (p + r)
    except ZeroDivisionError:
        f1 = 0

    return {
        'Precision': round(p, 5),
        'Recall': round(r, 5),
        'Accuracy': round(a, 5),
        'F1': round(f1, 5)
    }


def rescale2d(arr):
    m = _np.max(arr)
    n = _np.min(arr)
    return (arr - n) / (m - n)


def rescale3d(arrays):
    return list(rescale2d(arr) for arr in arrays)


def get_signed_diff_int8(image_arr1=None, image_arr2=None):
    signed_diff = _np.array(image_arr1 - image_arr2, dtype=_np.int8)
    fx = _np.array(signed_diff - _np.min(signed_diff), _np.uint8)
    fx = rescale2d(fx)
    return _np.array(fx * 255, _np.uint8)


def whiten_image2d(img_arr2d=None):
    img_arr2d = img_arr2d.copy()
    img_arr2d = (img_arr2d - img_arr2d.mean()) / img_arr2d.std()
    return _np.array(rescale2d(img_arr2d) * 255, dtype=_np.uint8)


def get_chunk_indexes(img_shape=(0, 0), chunk_shape=(0, 0), offset_row_col=None):
    """
    Returns a generator for four corners of each patch within image as specified.
    :param img_shape: Shape of the original image
    :param chunk_shape: Shape of desired patch
    :param offset_row_col: Offset for each patch on both x, y directions
    :return:
    """
    img_rows, img_cols = img_shape[:2]
    chunk_row, chunk_col = chunk_shape
    offset_row, offset_col = offset_row_col

    row_end = False
    for i in range(0, img_rows, offset_row):
        if row_end:
            continue
        row_from, row_to = i, i + chunk_row
        if row_to > img_rows:
            row_to = img_rows
            row_from = img_rows - chunk_row
            row_end = True

        col_end = False
        for j in range(0, img_cols, offset_col):
            if col_end:
                continue
            col_from, col_to = j, j + chunk_col
            if col_to > img_cols:
                col_to = img_cols
                col_from = img_cols - chunk_col
                col_end = True
            yield [int(row_from), int(row_to), int(col_from), int(col_to)]


def get_chunk_indices_by_index(img_shape=(0, 0), chunk_shape=(0, 0), indices=None):
    x, y = chunk_shape
    ix = []
    for (c1, c2) in indices:
        w, h = img_shape[:2]
        p, q, r, s = c1 - x // 2, c1 + x // 2, c2 - y // 2, c2 + y // 2
        if p < 0:
            p, q = 0, x
        if q > w:
            p, q = w - x, w
        if r < 0:
            r, s = 0, y
        if s > h:
            r, s = h - y, h
        ix.append([int(p), int(q), int(r), int(s)])
    return ix


def merge_patches(patches=None, image_size=(0, 0), patch_size=(0, 0), offset_row_col=None):
    """
    Merge different pieces of image to form a full image. Overlapped regions are averaged.
    :param patches: List of all patches to merge in order (left to right). (N * W * H * C)
    :param image_size: Full image size
    :param patch_size: A patch size(Patches must be uniform in size to be able to merge)
    :param offset_row_col: Offset used to chunk the patches.
    :return:
    """
    padded_sum = _np.zeros(image_size)
    non_zero_count = _np.zeros_like(padded_sum)
    for i, chunk_ix in enumerate(get_chunk_indexes(image_size, patch_size, offset_row_col)):
        row_from, row_to, col_from, col_to = chunk_ix

        patch = _np.array(patches[i]).squeeze()

        _pad = [(row_from, image_size[0] - row_to), (col_from, image_size[1] - col_to)]
        if len(image_size) == 3:
            _pad.append((0, 0))

        padded = _np.pad(patch, _pad, 'constant')
        padded_sum = padded + padded_sum
        non_zero_count = non_zero_count + _np.array(padded > 0).astype(int)
    non_zero_count[non_zero_count == 0] = 1
    return _np.array(padded_sum / non_zero_count, dtype=_np.uint8)


def expand_and_mirror_patch(full_img_shape=None, orig_patch_indices=None, expand_by=None):
    """
    Given a patch within an image, this function select a speciified region around it if present, else mirros it.
    It is useful in neuralnetworks like u-net which look for wide range of area than the actual input image.
    :param full_img_shape: Full image shape
    :param orig_patch_indices: Four cornets of the actual patch
    :param expand_by: Expand by (x, y ) in each dimension
    :return:
    """

    i, j = int(expand_by[0] / 2), int(expand_by[1] / 2)
    p, q, r, s = orig_patch_indices
    a, b, c, d = p - i, q + i, r - j, s + j
    pad_a, pad_b, pad_c, pad_d = [0] * 4
    if a < 0:
        pad_a = i - p
        a = 0
    if b > full_img_shape[0]:
        pad_b = b - full_img_shape[0]
        b = full_img_shape[0]
    if c < 0:
        pad_c = j - r
        c = 0
    if d > full_img_shape[1]:
        pad_d = d - full_img_shape[1]
        d = full_img_shape[1]
    return a, b, c, d, [(pad_a, pad_b), (pad_c, pad_d)]


def largest_cc(binary_arr=None):
    from skimage.measure import label
    labels = label(binary_arr)
    if labels.max() != 0:  # assume at least 1 CC
        largest = labels == _np.argmax(_np.bincount(labels.flat)[1:]) + 1
        return largest


def map_img_to_img2d(map_to, img):
    arr = map_to.copy()

    rgb = arr.copy()
    if len(arr.shape) == 2:
        rgb = _np.zeros((arr.shape[0], arr.shape[1], 3), dtype=_np.uint8)
        rgb[:, :, 0], rgb[:, :, 1], rgb[:, :, 2] = arr, arr, arr

    rgb[:, :, 0][img == 255] = 255
    rgb[:, :, 1][img == 255] = 0
    rgb[:, :, 2][img == 255] = 0
    return rgb


def remove_connected_comp(segmented_img, connected_comp_diam_limit=20):
    """
    Remove connected components of a binary image that are less than smaller than specified diameter.
    :param segmented_img: Binary image.
    :param connected_comp_diam_limit: Diameter limit
    :return:
    """

    from scipy.ndimage.measurements import label

    img = segmented_img.copy()
    structure = _np.ones((3, 3), dtype=_np.int)
    labeled, n_components = label(img, structure)
    for i in range(n_components):
        ixy = _np.array(list(zip(*_np.where(labeled == i))))
        x1, y1 = ixy[0]
        x2, y2 = ixy[-1]
        dst = _math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
        if dst < connected_comp_diam_limit:
            for u, v in ixy:
                img[u, v] = 0
    return img


def get_pix_neigh(i, j, eight=False):
    """
    Get four/ eight neighbors of an image.
    :param i: x position of pixel
    :param j: y position of pixel
    :param eight: Eight neighbors? Else four
    :return:
    """

    n1 = (i - 1, j - 1)
    n2 = (i - 1, j)
    n3 = (i - 1, j + 1)
    n4 = (i, j - 1)
    n5 = (i, j + 1)
    n6 = (i + 1, j - 1)
    n7 = (i + 1, j)
    n8 = (i + 1, j + 1)
    if eight:
        return [n1, n2, n3, n4, n5, n6, n7, n8]
    else:
        return [n2, n5, n7, n4]


def binarize(arr, thr=50, max=255):
    _arr = arr.copy()
    if _arr.max() > thr:
        _arr[_arr < thr] = 0
        _arr[_arr >= thr] = max

    return _arr


def masked_bboxcrop(arr, *apply_to, offset=21, threshold=10):
    """
    Binarize, mask, bbox crop image for largest connected component.
    """
    mask = arr.copy()
    if len(mask.shape) > 2:
        mask = mask[:, :, 1]

    mask[mask < threshold] = 0
    mask[mask >= threshold] = 255

    coords = _cv2.findNonZero(mask)
    x, y, w, h = _cv2.boundingRect(coords)
    a, b, c, d = y, y + h, x, x + w
    a, b, c, d = max(0, a - offset), min(arr.shape[0], b + offset), max(0, c - offset), min(arr.shape[1], d + offset)

    res = [arr[a:b, c:d]]

    for dat in apply_to:
        res.append(dat[a:b, c:d])
    res.append(mask[a:b, c:d])

    return res


def resize(array, size):
    img = _IMG.fromarray(array)
    down = any([a < b for a, b in zip(size, array.shape[:2])])
    array = _IMG.fromarray(array)

    if down:
        array.thumbnail(size, _IMG.ANTIALIAS)
    else:
        array = array.resize((int(size[0]), int(img.size[1] / img.size[0] * size[0])))
    return _np.array(array)


def _mean_std(file_path):
    arr = _np.array(_IMG.open(file_path))

    result = []

    try:
        if len(arr.shape) == 2:
            result.append([arr.mean(), arr.std()])
        elif len(arr.shape) > 2:
            mean = [arr[:, :, i].mean() for i in range(arr.shape[2])]
            std = [arr[:, :, i].std() for i in range(arr.shape[2])]
            result.append([mean, std])
    except:
        _tb.print_exc()
    return _np.array(result)


def mean_std(images_path=".", json_path=None, json_key='train', nw=4):
    if not json_path:
        files = [_os.path.join(images_path, f) for f in _os.listdir(images_path)]

    else:
        files = [_os.path.join(images_path, f) for f in _json.load(open(json_path))[json_key]]

    return _np.array(multiRun(nproc=nw, data_list=files, func=_mean_std)).mean(0) / 255.0



r"""
The main core of EasyTorch
"""

import math as _math
import os as _os

import torch as _torch

import easytorch.utils as _etutils
from easytorch.config.state import *
from easytorch.metrics import metrics as _metrics
from easytorch.utils.logger import *
from easytorch.utils.tensorutils import initialize_weights as _init_weights
from .vision import plotter as _log_utils
import traceback as _tb

_sep = _os.sep


class ETTrainer:
    def __init__(self, args=None, data_handle=None, **kw):
        r"""
        args: receives the arguments passed by the ArgsParser.
        cache: Initialize all immediate things here. Like scores, loss, accuracies...
        nn:  Initialize our models here.
        optimizer: Initialize our optimizers.
        """
        self.cache = {}
        self.args = _etutils.FrozenDict(args)
        self.data_handle = data_handle

        self.nn = {}
        self.optimizer = {}
        self.device = {'gpu': args.get('gpu', 'cpu')}

    def init_nn(self,
                init_models=True,
                init_weights=True,
                init_optimizer=True,
                set_device=True):
        r"""
        Call to user implementation of:
            Initialize models.
            Initialize random/pre-trained weights.
            Initialize/Detect GPUS.
            Initialize optimizer.
        """

        if init_models: self._init_nn_model()
        # Print number of parameters in all models.
        if init_models and self.args['verbose']:
            for k, m in [(_k, _m) for _k, _m in self.nn.items() if isinstance(_m, _torch.nn.Module)]:
                success(f'Total params in {k}:' f' {sum(p.numel() for p in m.parameters() if p.requires_grad)}')

        if init_optimizer: self._init_optimizer()
        if init_weights: self._init_nn_weights()
        if set_device: self._set_device()

    def _init_nn_weights(self):
        r"""
        By default, will initialize network with Kaimming initialization.
        If path to pretrained weights are given, it will be used instead.
        """
        if self.args['pretrained_path'] is not None:
            self.load_checkpoint(self.args['pretrained_path'],
                                 self.args.get('load_model_state', True),
                                 self.args.get('load_optimizer_state', False))

        elif self.args['phase'] == 'train':
            _torch.manual_seed(self.args['seed'])
            for mk in self.nn:
                _init_weights(self.nn[mk])

    def load_checkpoint(self,
                        full_path,
                        load_model_state=True,
                        load_optimizer_state=True,
                        src=MYSELF,
                        map_location=_torch.device('cpu')):
        r"""
        Load checkpoint from the given path:
            If it is an easytorch checkpoint, try loading all the models.
            If it is not, assume it's weights to a single model and laod to first model.
        """
        chk = _torch.load(full_path, map_location=map_location)
        if chk.get('_its_origin_', 'Unknown').lower() == src:
            if load_model_state:
                for m in chk['models']:
                    try:
                        self.nn[m].module.load_state_dict(chk['models'][m])
                    except:
                        self.nn[m].load_state_dict(chk['models'][m])

            if load_optimizer_state:
                for m in chk['optimizers']:
                    try:
                        self.optimizer[m].module.load_state_dict(chk['optimizers'][m])
                    except:
                        self.optimizer[m].load_state_dict(chk['optimizers'][m])
        else:
            mkey = list(self.nn.keys())[0]
            try:
                self.nn[mkey].module.load_state_dict(chk)
            except:
                self.nn[mkey].load_state_dict(chk)

    def _init_nn_model(self):
        r"""
        User cam override and initialize required models in self.nn dict.
        """
        raise NotImplementedError('Must be implemented in child class.')

    def _set_device(self):
        r"""
        Initialize GPUs based on whats provided in args(Default [0])
        Expects list of GPUS as [0, 1, 2, 3]., list of GPUS will make it use DataParallel.
        If no GPU is present, CPU is used.
        """
        if self.args.get('use_ddp'):
            _device_ids = []
            if self.args['gpu'] is not None:
                _device_ids.append(self.device['gpu'])

            for model_key in self.nn:
                self.nn[model_key] = self.nn[model_key].to(self.device['gpu'])
            for model_key in self.nn:
                self.nn[model_key] = _torch.nn.parallel.DistributedDataParallel(self.nn[model_key],
                                                                                device_ids=_device_ids)
        elif len(self.args['gpus']) >= 1:
            self.device['gpu'] = _torch.device(f"cuda:{self.args['gpus'][0]}")
            if len(self.args['gpus']) >= 2:
                for model_key in self.nn:
                    self.nn[model_key] = _torch.nn.DataParallel(self.nn[model_key], self.args['gpus'])
            for model_key in self.nn:
                self.nn[model_key] = self.nn[model_key].to(self.device['gpu'])

    def _init_optimizer(self):
        r"""
        Initialize required optimizers here. Default is Adam,
        """
        first_model = list(self.nn.keys())[0]
        self.optimizer['adam'] = _torch.optim.Adam(self.nn[first_model].parameters(),
                                                   lr=self.args['learning_rate'])

    def new_meter(self):
        r"""
        User can override to supply desired implementation of easytorch.metrics.ETMetrics().
            Example: easytorch.metrics.Pr11a() will work with precision, recall, F1, Accuracy, IOU scores.
        """
        return _metrics.ETMeter(num_averages=1)

    def save_checkpoint(self,
                        full_path,
                        save_model_state=True,
                        save_optimizer_state=True,
                        src=MYSELF):

        checkpoint = {'_its_origin_': src}
        if save_model_state:
            checkpoint['models'] = {}
            for k in self.nn:
                try:
                    checkpoint['models'][k] = self.nn[k].module.state_dict()
                except:
                    checkpoint['models'][k] = self.nn[k].state_dict()

        if save_optimizer_state:
            checkpoint['optimizers'] = {}
            for k in self.optimizer:
                try:
                    checkpoint['optimizers'][k] = self.optimizer[k].module.state_dict()
                except:
                    checkpoint['optimizers'][k] = self.optimizer[k].state_dict()
        _torch.save(checkpoint, full_path)

    def init_experiment_cache(self):
        r"""What scores you want to plot."""
        self.cache['log_header'] = 'Loss,Accuracy'

        r"""This is for best model selection: """
        r"""It tells which metrics to monitor and either to maximize(F1 score), minimize(MSE)"""
        self.cache.update(monitor_metric='time', metric_direction='maximize')

    def iteration(self, batch) -> dict:
        r"""
        Note: 'loss' adn 'meter' keys are required in the return dict, whereas others are optional
        """
        return {}

    def save_predictions(self, dataset, its):
        r"""
        If one needs to save complex predictions result like predicted segmentations.
         -Especially with U-Net architectures, we split images and train.
        Once the argument --sp/-sparse-load is set to True,
        the argument 'its' will receive all the patches of single image at a time.
        From there, we can recreate the whole image.
        """
        return None

    def evaluation(self,
                   epoch=1,
                   mode='eval',
                   dataloaders: list = None,
                   save_pred=False) -> _metrics.ETMeter:
        for k in self.nn:
            self.nn[k].eval()

        eval_meter = self.new_meter()

        if not dataloaders:
            return eval_meter

        info(f'{mode} ...', self.args['verbose'])

        def _update_scores(_out, _it, _meter):
            if isinstance(_out, _metrics.ETMeter):
                _meter.accumulate(_out)
            else:
                _meter.accumulate(_it['meter'])

        with _torch.no_grad():
            for loader in dataloaders:
                try:
                    its = []
                    meter = self.new_meter()

                    for i, batch in enumerate(loader, 1):
                        it = self.iteration(batch)

                        if save_pred:
                            if self.args['load_sparse']:
                                its.append(it)
                            else:
                                _update_scores(self.save_predictions(loader.dataset, it), it, meter)
                        else:
                            _update_scores(None, it, meter)

                        if self.args['verbose'] and len(dataloaders) <= 1 and lazy_debug(i, add=epoch):
                            info(f"  Itr:{i}/{len(loader)}, {it['meter']}")

                    if save_pred and self.args['load_sparse']:
                        its = self._reduce_iteration(its)
                        _update_scores(self.save_predictions(loader.dataset, its), its, meter)

                    if self.args['verbose'] and len(dataloaders) > 1:
                        info(f" {mode}, {meter}")

                    eval_meter.accumulate(meter)
                except:
                    _tb.print_exc()

        info(f"{self.cache['experiment_id']} {mode} {eval_meter.get()}", self.args['verbose'])
        return eval_meter

    def _reduce_iteration(self, its) -> dict:
        reduced = {}.fromkeys(its[0].keys(), None)
        for key in reduced:
            if isinstance(its[0][key], _metrics.ETMeter):
                reduced[key] = self.new_meter()
                [reduced[key].accumulate(ik[key]) for ik in its]
            else:
                def collect(k=key, src=its):
                    _data = []
                    is_tensor = isinstance(src[0][k], _torch.Tensor)
                    is_tensor = is_tensor and not src[0][k].requires_grad and src[0][k].is_leaf
                    for ik in src:
                        if is_tensor:
                            _data.append(ik[k] if len(ik[k].shape) > 0 else ik[k].unsqueeze(0))
                        else:
                            _data.append(ik[k])
                    if is_tensor:
                        return _torch.cat(_data)
                    return _data

                reduced[key] = collect

        return reduced

    def _on_iteration_end(self, **kw):
        r"""
        Any logic to run after an iteration ends.
        """
        pass

    def _check_validation_score(self, epoch, training_meter, validation_meter):
        r"""
        Save the current model as best if it has better validation scores.
        """
        self.cache['_monitored_metrics_key_'], sc = validation_meter.extract(self.cache['monitor_metric'])
        improved = False
        if self.cache['metric_direction'] == 'maximize':
            improved = sc > self.cache['best_val_score'] + self.args.get('score_delta', SCORE_DELTA)
        elif self.cache['metric_direction'] == 'minimize':
            improved = sc < self.cache['best_val_score'] - self.args.get('score_delta', SCORE_DELTA)
        return {'improved': improved, 'score': sc}

    def _stop_early(self, **kw):
        r"""
        Stop the training based on some criteria.
         For example: the implementation below will stop training if the validation
         scores does not improve within a 'patience' number of epochs.
        """
        if self.args['patience'] and kw['epoch'] - self.cache['best_val_epoch'] >= self.args['patience']:
            return True

        if self.cache['metric_direction'] == 'maximize':
            return self.cache['best_val_score'] == self.args.get('score_max', SCORE_MAX)
        elif self.cache['metric_direction'] == 'minimize':
            return self.cache['best_val_score'] == self.args.get('score_min', SCORE_MIN)

        return False

    def _save_progress(self, epoch):
        _log_utils.plot_progress(self.cache, experiment_id=self.cache['experiment_id'],
                                 plot_keys=[LogKey.TRAIN_LOG, LogKey.VALIDATION_LOG],
                                 epoch=epoch)

    def training_iteration(self, i, batch) -> dict:
        r"""
        Learning step for one batch.
        We decoupled it so that user could implement any complex/multi/alternate training strategies.
        """
        it = self.iteration(batch)
        it['loss'].backward()
        if i % self.args.get('grad_accum_iters', 1) == 0:
            for optim in self.optimizer:
                self.optimizer[optim].step()
                self.optimizer[optim].zero_grad()
        return it

    def reduce_scores(self, accumulator: list, distributed=False) -> _metrics.ETMeter:
        meter = self.new_meter()
        if all([a is None for a in accumulator]):
            return meter

        for m in accumulator:
            meter.accumulate(m)

        if distributed:
            meter.averages.dist_gather(device=self.device['gpu'])
            for mk in meter.metrics:
                meter.metrics[mk].dist_gather(device=self.device['gpu'])

        return meter

    def save_if_better(self, epoch, training_meter=None, validation_meter=None):
        val_check = self._check_validation_score(epoch, training_meter, validation_meter)
        if val_check['improved']:
            self.save_checkpoint(self.cache['log_dir'] + _sep + self.cache['best_checkpoint'])
            self.cache['best_val_score'] = val_check['score']
            self.cache['best_val_epoch'] = epoch
            success(f" *** Best Model Saved!!! *** : {self.cache['best_val_score']}", self.args['verbose'])
        else:
            info(
                f"Not best: {val_check['score']}, {self.cache['best_val_score']} in ep: {self.cache['best_val_epoch']}",
                self.args['verbose'])

    def _global_debug(self, running_meter, **kw):
        """Update running accumulators."""
        running_meter.accumulate(kw['meter'])

        """ Reset iteration accumulator """
        N = kw['num_iters']
        i, e = kw['i'], kw['epoch']

        if lazy_debug(i, add=e + 1) or i == N:
            info(f"Ep:{e}/{self.args['epochs']}, Itr:{i}/{N}, {running_meter}", self.args['verbose'])
            r"""Debug and reset running accumulators"""

            if not self.args['use_ddp']:
                """Plot only in non-ddp mode to maintain consistency"""
                self.cache[LogKey.TRAIN_LOG].append(running_meter.get())

            running_meter.reset()

    def inference(self, mode='test', save_predictions=True, datasets: list = None, distributed=False):
        if not isinstance(datasets, list):
            datasets = [datasets]

        loaders = []
        for d in [_d for _d in datasets if _d]:
            loaders.append(
                self.data_handle.get_loader(
                    handle_key=mode, shuffle=False, dataset=d, distributed=distributed
                )
            )

        return self.evaluation(
            mode=mode,
            dataloaders=[_l for _l in loaders if _l],
            save_pred=save_predictions
        )

    def train(self, train_dataset, validation_dataset) -> None:
        info('Training ...', self.args['verbose'])

        train_loader = self.data_handle.get_loader(
            handle_key='train',
            shuffle=True,
            dataset=train_dataset,
            distributed=self.args['use_ddp']
        )

        val_loader = self.data_handle.get_loader(
            handle_key='validation',
            shuffle=False,
            dataset=validation_dataset,
            distributed=self.args['use_ddp'] and self.args.get('distributed_validation'),
            use_unpadded_sampler=True
        )

        if val_loader is not None and not isinstance(val_loader, list):
            val_loader = [val_loader]

        for ep in range(1, self.args['epochs'] + 1):
            for k in self.nn:
                self.nn[k].train()

            """Collect accumulated iterations data"""
            its = []

            """Collect epoch metrics and averages"""
            epoch_meter = self.new_meter()

            """Keep track of running metrics and averages for logging/plotting"""
            _meter = self.new_meter()

            if self.args.get('use_ddp'):
                train_loader.sampler.set_epoch(ep)

            num_iters = len(train_loader) // self.args['grad_accum_iters']
            for i, batch in enumerate(train_loader, 1):
                its.append(self.training_iteration(i, batch))
                """When end of iteration"""
                if i % self.args['grad_accum_iters'] == 0:
                    it = self._reduce_iteration(its)

                    """Update global accumulators"""
                    its = []
                    it['num_iters'] = num_iters
                    it['i'] = i // self.args['grad_accum_iters']
                    epoch_meter.accumulate(it['meter'])

                    if self.args['is_master']:
                        self._global_debug(_meter, epoch=ep, **it)
                    self._on_iteration_end(i=i, epoch=ep, it=it)

            epoch_meter = self.reduce_scores(
                [epoch_meter],
                distributed=self.args['use_ddp']
            )

            epoch_details = {
                'epoch': ep,
                'training_meter': epoch_meter,
                "validation_meter": None
            }

            """Validation step"""
            if val_loader is not None:
                val_out = self.evaluation(ep, mode='validation', dataloaders=val_loader, save_pred=False)
                epoch_details['validation_meter'] = self.reduce_scores(
                    [val_out],
                    distributed=self.args['use_ddp'] and self.args.get('distributed_validation')
                )

            info('--', self.args['is_master'])
            self._on_epoch_end(**epoch_details)
            if self.args['is_master']:
                self._global_epoch_end(**epoch_details)

            if self._stop_early(**epoch_details):
                break

        """Plot at the end regardless."""
        self._save_progress(epoch=ep)

    def _global_epoch_end(self, epoch, training_meter=None, validation_meter=None):
        if training_meter:
            self.cache[LogKey.TRAIN_LOG].append(training_meter.get())

        if validation_meter:
            self.save_if_better(epoch, training_meter, validation_meter)
            self.cache[LogKey.VALIDATION_LOG].append(validation_meter.get())

        if lazy_debug(epoch, _math.log(epoch)):
            self._save_progress(epoch=epoch)

    def _on_epoch_end(self, epoch, training_meter=None, validation_meter=None):
        """Local epoch end"""
        pass



import argparse as _ap
from collections import OrderedDict as _ODict
from easytorch.config.state import *


def boolean_string(s):
    try:
        return str(s).strip().lower() == 'true'
    except:
        return False


class StoreDictKeyPairSS(_ap.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        my_dict = _ODict()
        for kv in values.split(","):
            k, v = kv.split("=")
            my_dict[k] = v
        setattr(namespace, self.dest, my_dict)


default_ap = _ap.ArgumentParser()

default_ap.add_argument('-ph', '--phase', default=None, choices=['train', 'test'], type=str,
                        help='Can be train/test.')
default_ap.add_argument("-b", "--batch_size", default=4, type=int, help="Mini-batch size.")
default_ap.add_argument('-ep', '--epochs', default=11, type=int, help='Number of epochs.')
default_ap.add_argument('-gi', '--grad_accum_iters', default=1, type=int,
                        help='Number of iterations for gradient accumulation.')
default_ap.add_argument('-lr', '--learning_rate', default=0.001, type=float, help='Learning rate.')
default_ap.add_argument('-gpus', '--gpus', default=list(range(NUM_GPUS)), nargs='*', type=int,
                        help='How many gpus to use?')
default_ap.add_argument('-pin', '--pin_memory', default=False, type=boolean_string, help='Pin Memory.')
default_ap.add_argument('-nw', '--num_workers', default=0, type=int,
                        help='Number of workers to work on data loading.')
default_ap.add_argument('-data', '--dataset_dir', default='', type=str, help='Root path to Datasets.')
default_ap.add_argument('-lim', '--load_limit', default=None, type=int, help='Data load limit')
default_ap.add_argument('-log', '--log_dir', default='net_logs', type=str, help='Logging directory.')
default_ap.add_argument('-pt', '--pretrained_path', default=None, type=str,
                        help='Full path to pretrained weights(It will be loaded before training.)')
default_ap.add_argument('-v', '--verbose', default=True, type=boolean_string,
                        help='Prints information on different steps.')
default_ap.add_argument('-seed-all', '--seed_all', default=False, type=boolean_string, help='Seed everything.')
default_ap.add_argument('-seed', '--seed', default=None, type=int, help='Seed everything.')
default_ap.add_argument('-f', '--force', default=False, type=boolean_string, help='Force')
default_ap.add_argument('-pat', '--patience', default=None, type=int, help='Early Stopping patience epochs.')
default_ap.add_argument('-lsp', '--load_sparse', default=False, type=boolean_string, help='Load sparse dataset.')
default_ap.add_argument('-nf', '--num_folds', default=None, type=int, help='Number of folds.')
default_ap.add_argument('-spl', '--split_ratio', default=None, nargs='*', type=float,
                        help='Split ratio. Eg: 0.6 0.2 0.2 or 0.8 0.2. Exclusive to num_fold.')

default_ap.add_argument('-ddp', '--use_ddp', default=False, type=boolean_string, help='Use DDP?')
default_ap.add_argument('--multi-load', default=True, type=boolean_string, help='Multi loading data.')


_im, _ = default_ap.parse_known_args()
_args = vars(_im)
if _args.get('use_ddp'):
    default_ap.add_argument('--node-rank', default=0, type=int,
                            help='Node rank for distributed training')
    default_ap.add_argument('--num-nodes', default=1, type=int,
                            help='Number of nodes for distributed training')
    default_ap.add_argument('--world-size', default=None, type=int,
                            help='World size(Total participating processes(GPUs))')
    default_ap.add_argument('--init-method', default='env://', type=str,
                            help='Initialization method used to start distributed training.')
    default_ap.add_argument('--dist-backend', default='gloo', type=str,
                            help='Distributed backend')

    default_ap.add_argument('--master-addr', default='127.0.0.1', type=str,
                            help='Master node address.')
    default_ap.add_argument('--master-port', default='8998', type=str,
                            help='Master node address.')

_known, _unknown = default_ap.parse_known_args()
default_args = vars(_known)



import os as _os

import torch as _torch
import torch.distributed as _dist
import torch.multiprocessing as _mp
import torch.nn as _nn
import torch.optim as _optim
from easytorch import default_args as _args


def _setup(rank, args):
    if not args.get('world_size'):
        args['world_size'] = args['num_gpus'] * args['num_nodes']

    world_rank = args['node_rank'] * args['num_gpus'] + rank
    _dist.init_process_group(backend=args['dist_backend'],
                             init_method=args['init_method'],
                             rank=world_rank, world_size=args['world_size'])

    total_ranks = _torch.Tensor([1]).to(args['gpus'][rank])
    _dist.all_reduce(total_ranks, _dist.ReduceOp.SUM)
    if rank == 0:
        print("\n\t***** DDP injection successful!!! *****")
        print(f"\tTotal nodes: {args['num_nodes']}")
        print(f'\tTotal participating processes: {int(total_ranks.item())}')
        print("\t***************************************")


def _train_distributed(rank, args, data_loaders, model):
    print(f'Rank {rank} initializing...')
    gpu = args['gpus'][rank]
    _setup(rank, args)

    loss_fn = _nn.MSELoss()
    model = model.to(gpu)
    optimizer = _optim.SGD(model.parameters(), lr=0.001)
    for batch in data_loaders['train']:
        data = batch[0].to(gpu)
        labels = batch[1].to(gpu)
        outputs = model(data)

        loss = loss_fn(outputs, labels).backward()
        loss.backwards()

        optimizer.step()
        optimizer.zero_grad()

        """Sync grads across (Best way is to wrap model with DistributedParallel instead)"""
        for name, params in model.named_parameters():
            _dist.all_reduce(params.grad.data, op=_dist.ReduceOp.SUM)
            params.grad.data /= float(args['world_size'])


def _run_distributed(args, model, data_loaders: dict):
    if args['num_gpus'] > 0:
        args['gpus'] = list(range(args['num_gpus']))
    elif args['world_size'] is not None:
        args['gpus'] = [None] * args['world_size']
    else:
        raise ValueError(
            'Something is wrong! Either run in a machine with GPU '
            'or provide world_size value and gloo backend for CPU usages.'
        )

    _os.environ['MASTER_ADDR'] = args['master_addr']
    _os.environ['MASTER_PORT'] = args['master_port']
    _mp.spawn(_train_distributed, nprocs=len(args['gpus']), args=(args, data_loaders, model))


"""
Run in cpu
init_method = 'gloo'
world_size = <desired number of processes>
"""

"""
Run in GPUs
init_method = 'nccl'
world_size = None, since automatically determined by number of GPUs.
"""

if __name__ == "__main__":
    _args['num_gpus'] = len(_args['gpus'])
    data_loaders = {
        'train': [],
        'validation': [],
        'test': []
    }
    model = _nn.Linear(10, 10)
    _run_distributed(_args, model, data_loaders)



import math as _math
import os as _os

import matplotlib.pyplot as _plt
import seaborn as _sns

import numpy as _np
import pandas as _pd

_plt.switch_backend('agg')
_plt.rcParams["figure.figsize"] = [16, 9]
_sns.set()

COLORS = ['blue', 'maroon', 'magenta', 'teal', 'red', 'blueviolet', 'brown', 'cadetblue',
          'chartreuse', 'coral', 'darkslateblue', 'cornflowerblue', 'indigo', 'black', 'cyan', 'navy']


def plot_progress(cache, experiment_id='', plot_keys=[], num_points=31, epoch=None):
    r"""
    Custom plot to plot data from the cache by keys.
    """
    for k in plot_keys:
        D = _np.array(cache.get(k, []))
        if len(D) == 0 or cache.get('log_header') is None:
            continue

        i = 0
        for plot_id, header in enumerate(cache['log_header'].split('|')):
            _plt.clf()
            header = header.split(',')
            j = i + len(header)
            data = D[:, i:j]
            if _np.sum(data) <= 0:
                continue

            df = _pd.DataFrame(data, columns=header)
            if len(df) == 0:
                continue

            color = COLORS[i:j]

            rollin_window = max(df.shape[0] // num_points, 3)
            ax = df.plot(x_compat=True, alpha=0.11, legend=0, color=color)

            rolling = df.rolling(rollin_window, min_periods=1).mean()
            rolling.plot(ax=ax, title=k.upper(), color=color)

            if epoch and epoch != df.shape[0]:
                """
                Set correct epoch as x-tick-labels.
                """
                xticks = list(range(0, df.shape[0], df.shape[0] // epoch)) + [df.shape[0] - 1]
                step = int(_math.log(len(xticks) + 1) + len(xticks) // num_points + 1)
                xticks_range = list(range(len(xticks)))[::step]
                xticks = xticks[::step]
                ax.set_xticks(xticks)
                ax.set_xticklabels(xticks_range)

            _plt.xlabel('Epochs')
            _plt.savefig(cache['log_dir'] + _os.sep + f"{experiment_id}_{k}_{plot_id}.png", bbox_inches='tight')
            _plt.close('all')
            i = j



import copy as _copy
import json as _json
import os as _os
import random as _rd

import numpy as _np

_sep = _os.sep


def create_ratio_split(files, save_to_dir=None, ratio: list = None, first_key='train', name='SPLIT',
                       shuffle_files=True):
    if shuffle_files:
        _rd.seed(len(files))
        _rd.shuffle(files)

    keys = [first_key]
    if len(ratio) == 2:
        keys.append('test')
    elif len(ratio) == 3:
        keys.append('validation')
        keys.append('test')

    _ratio = ratio[::-1]
    locs = _np.array([sum(_ratio[0:i + 1]) for i in range(len(ratio) - 1)])
    locs = (locs * len(files)).astype(int)
    splits = _np.split(files[::-1], locs)[::-1]
    splits = dict([(k, sp.tolist()[::-1]) for k, sp in zip(keys, splits)])
    if save_to_dir:
        f = open(save_to_dir + _sep + f'{name}.json', "w")
        f.write(_json.dumps(splits))
        f.close()
    else:
        return splits


def create_k_fold_splits(files, k=0, save_to_dir=None, shuffle_files=True, name='SPLIT'):
    if shuffle_files:
        _rd.seed(len(files))
        _rd.shuffle(files)

    file_ix = _np.arange(len(files))
    ix_splits = _np.array_split(file_ix, k)
    for i in range(len(ix_splits)):

        test_ix = ix_splits[i].tolist()
        val_ix = ix_splits[(i + 1) % len(ix_splits)].tolist()
        train_ix = _np.delete(file_ix.copy(), _np.array(test_ix + val_ix))

        splits = {"train": [files[ix] for ix in train_ix],
                  "validation": [files[ix] for ix in val_ix],
                  "test": [files[ix] for ix in test_ix]}

        if save_to_dir:
            f = open(save_to_dir + _sep + f"{name}_{i}.json", "w")
            f.write(_json.dumps(splits))
            f.close()
        else:
            return splits


def uniform_mix_two_lists(smaller, larger, shuffle=True):
    if shuffle:
        _rd.shuffle(smaller)
        _rd.shuffle(larger)

    len_smaller, len_larger = len(smaller), len(larger)

    accumulator = []
    while len(accumulator) < len_smaller + len_larger:
        try:
            for i in range(int(len_larger / len_smaller)):
                accumulator.append(larger.pop())
        except Exception:
            pass
        try:
            accumulator.append(smaller.pop())
        except Exception:
            pass

    return accumulator


def make_weights_for_balanced_classes(images, nclasses):
    count = [0] * nclasses
    for item in images:
        count[item[1]] += 1
    weight_per_class = [0.] * nclasses
    N = float(sum(count))
    for i in range(nclasses):
        weight_per_class[i] = N / float(count[i])
    weight = [0] * len(images)
    for idx, val in enumerate(images):
        weight[idx] = weight_per_class[val[1]]
    return weight


def should_create_splits_(log_dir, dspec, args):
    if dspec['split_dir'].endswith('.json') or dspec['split_dir'].endswith('.txt'):
        return False

    if dspec.get('split_dir') and _os.path.exists(dspec.get('split_dir')) and len(list(
            _os.listdir(dspec.get('split_dir')))) > 0:
        return False

    dspec['split_dir'] = log_dir + _sep + 'splits'
    if _os.path.exists(dspec['split_dir']) and len(list(_os.listdir(dspec['split_dir']))) > 0:
        return False

    _os.makedirs(dspec['split_dir'], exist_ok=True)
    if args['num_folds'] is None and args['split_ratio'] is None:
        with open(dspec['split_dir'] + _sep + 'experiment.json', 'w') as sp:
            sp.write(_json.dumps({'train': [], 'validation': [], 'test': []}))
        return False

    return True


def default_data_splitter_(files, dspec, args):
    r"""
    Initialize k-folds for given dataspec.
        If: custom splits path is given it will use the splits from there
        else: will create new k-splits and run k-fold cross validation.
    """
    if args.get('num_folds') is not None:
        create_k_fold_splits(
            files=files,
            k=args['num_folds'],
            save_to_dir=dspec['split_dir'],
            name=dspec['name']
        )
    elif args.get('split_ratio') is not None:
        create_ratio_split(
            files=files,
            save_to_dir=dspec['split_dir'],
            ratio=args['split_ratio'],
            name=dspec['name']
        )



import math as _math


def error(msg, debug=True):
    if debug:
        print(f"[ERROR]! {msg}")


def warn(msg, debug=True):
    if debug:
        print(f"[WARNING]! {msg}")


def info(msg, debug=True):
    if debug: print(f"{msg}")


def success(msg, debug=True):
    if debug:
        print(f"[SUCCESS]! {msg}")


def lazy_debug(x, add=0):
    _scale = int(_math.log(max(x, 1)) * _math.log(max(add, 1)))
    return x % (_scale + 1) == 0



import json as _json
import os as _os
import pprint as _pp
import random as _random
import typing
from argparse import ArgumentParser as _AP

import numpy as _np
import torch as _torch
import torch.distributed as _dist
import torch.multiprocessing as _mp

import easytorch.config as _conf
import easytorch.utils as _utils
from easytorch.config.state import *
from easytorch.data import ETDataset, ETDataHandle, multiproc as _mproc, DiskCache as _DiskCache
from easytorch.trainer import ETTrainer
from easytorch.utils.logger import *
import uuid as _uuid
from datetime import datetime as _dtime
from pathlib import Path as _Path

_sep = _os.sep


def _ddp_worker(rank, self, trainer_cls, dataset_cls, data_handle_cls, is_pooled):
    self.args['gpu'] = self.args['gpus'][rank]
    self.args['verbose'] = rank == MASTER_RANK
    world_size = self.args['world_size']
    if not world_size:
        world_size = self.args['num_gpus'] * self.args['num_nodes']
    world_rank = self.args['node_rank'] * self.args['num_gpus'] + rank

    self.args['is_master'] = world_rank == MASTER_RANK
    _dist.init_process_group(backend=self.args['dist_backend'],
                             init_method=self.args['init_method'],
                             world_size=world_size, rank=world_rank)
    if is_pooled:
        self._run_pooled(trainer_cls, dataset_cls, data_handle_cls)
    else:
        self._run(trainer_cls, dataset_cls, data_handle_cls)


def _cleanup(trainer):
    for data_handle_key in trainer.data_handle.datasets:
        if hasattr(trainer.data_handle.datasets[data_handle_key], 'diskcache') and isinstance(
                trainer.data_handle.datasets[data_handle_key].diskcache, _DiskCache):
            trainer.data_handle.datasets[data_handle_key].diskcache.clear()


class EasyTorch:
    _MODES_ = [Phase.TRAIN, Phase.TEST]
    _MODE_ERR_ = \
        "####  [ERROR]  ### argument 'phase' is required and must be passed to either" \
        '\n\t1). EasyTorch(..,phase=<value>,..)' \
        '\n\t2). runtime arguments 2). python main.py -ph <value> ...' \
        f'\nPossible values are:{_MODES_}'

    def __init__(self, dataspecs: typing.List[dict] = None,
                 args: typing.Union[dict, _AP] = _conf.default_args,
                 phase: str = _conf.default_args['phase'],
                 batch_size: int = _conf.default_args['batch_size'],
                 grad_accum_iters: int = _conf.default_args['grad_accum_iters'],
                 epochs: int = _conf.default_args['epochs'],
                 learning_rate: float = _conf.default_args['learning_rate'],
                 gpus: typing.List[int] = _conf.default_args['gpus'],
                 pin_memory: bool = _conf.default_args['pin_memory'],
                 num_workers: int = _conf.default_args['num_workers'],
                 dataset_dir: str = _conf.default_args['dataset_dir'],
                 load_limit: int = _conf.default_args['load_limit'],
                 log_dir: str = _conf.default_args['log_dir'],
                 pretrained_path: str = _conf.default_args['pretrained_path'],
                 verbose: bool = _conf.default_args['verbose'],
                 seed_all: int = _conf.default_args['seed_all'],
                 seed: int = _conf.default_args['seed'],
                 force: bool = _conf.default_args['force'],
                 patience: int = _conf.default_args['patience'],
                 load_sparse: bool = _conf.default_args['load_sparse'],
                 num_folds=_conf.default_args['num_folds'],
                 split_ratio=_conf.default_args['split_ratio'],
                 dataloader_args: dict = None,
                 **kw):
        """
        Order of precedence of arguments is(Higher will override the lower):
            1. Default args as in easytorch.conf.default_args
            2. Listed args in __init__ method
            3. kwargs in **kw
        @param dataspecs: List of dict with which dataset details like data_files path, ground truth path...
                Example: [{'data_dir':'images', 'labels_dir':'manuals', 'splits_dir':'splits'}]
                Each key with _dir in it will be appended before the value provided in 'dataset_dir' argument.
        @param args: An argument parser, or, dict. (Defaults are loaded from easytorch.conf.default_args.)
                    Note: values in args will be overridden by the listed args below if provided.
        @param phase: phase of operation; train/test. (Default: None)
                    train phase will run all train, validation, and test step.
        @param batch_size: Default is 32
        @param grad_accum_iters: Number of iterations to accumulate gradients. (Default 1)
        @param epochs: Default is 21
        @param learning_rate: Default is 0.001
        @param gpus: Default [0]. But set to [](or cpu) if no gpus found.
        @param pin_memory: Default is True if cuda found.
        @param num_workers: Default is 4.
        @param dataset_dir: Default is ''. Path to some dataset folder.
        @param load_limit: Load limit for data items for debugging pipeline with few data sample. Default is 1e11
        @param log_dir: Directory path to place all saved models, plots.... Default is net_logs/
        @param pretrained_path: Path to load pretrained model. Default is None
        @param verbose: Show logs? Default is True
        @param seed_all: If seeds to use for reproducibility. Default is False.
        @param force: Force to clear previous logs in log_dir(if any).
        @param patience: Set patience epochs to stop training. Uses validation scores. Default is 11.
        @param load_sparse: Loads test dataset in single data loader to recreate data(eg images) from prediction. Default is False.
        @param num_folds: Number of k-folds to split the data(eg folder with images) into. Default is None.
                        However, if a custom json split(s) are provided with keys train, validation,
                        test is provided in split_dir folder as specified in dataspecs, it will be loaded.
        @param split_ratio: Ratio to split files as specified in data_dir in dataspecs into. Default is 0.6, 0.2. 0.2.
                        However, if a custom json split(s) are provided with keys train, validation,
                        test is provided in split_dir folder as specified in dataspecs, it will be loaded.
        @param dataloader_args: dict with keys train, test, and validation that will ovveride corresponding dataloader args.
                For example, different batch size for validation loader.
        @param kw: Extra kwargs.
        """
        self._init_args(args)

        self.args.update(phase=phase)
        self.args.update(batch_size=batch_size)
        self.args.update(grad_accum_iters=grad_accum_iters)
        self.args.update(epochs=epochs)
        self.args.update(learning_rate=learning_rate)
        self.args.update(gpus=gpus)
        self.args.update(pin_memory=pin_memory)
        self.args.update(num_workers=num_workers)
        self.args.update(dataset_dir=dataset_dir)
        self.args.update(load_limit=load_limit)
        self.args.update(log_dir=log_dir)
        self.args.update(pretrained_path=pretrained_path)
        self.args.update(verbose=verbose)
        self.args.update(seed_all=seed_all)
        self.args.update(seed=seed)
        self.args.update(force=force)
        self.args.update(patience=patience)
        self.args.update(load_sparse=load_sparse)
        self.args.update(num_folds=num_folds)
        self.args.update(split_ratio=split_ratio)
        self.args.update(**kw)

        self.dataloader_args = dataloader_args if dataloader_args else {}
        assert (self.args.get('phase') in self._MODES_), self._MODE_ERR_

        self._init_dataspecs(dataspecs)

        self._device_check()
        self._ddp_setup()
        self._make_reproducible()
        self.args.update(is_master=self.args.get('is_master', True))
        self.args['RUN-ID'] = _dtime.now().strftime("ET-%Y%m%d-%H%M%S-") + _uuid.uuid4().hex[:8].upper()
        self.args['log_dir'] = self.args['log_dir'] + _sep + self.args['phase'].upper()

    def _device_check(self):
        self.args['gpus'] = self.args['gpus'] if self.args.get('gpus') else []
        if self.args['verbose'] and len(self.args['gpus']) > NUM_GPUS:
            warn(f"{len(self.args['gpus'])} GPU(s) requested "
                 f"but {NUM_GPUS if CUDA_AVAILABLE else 'GPU(s) not'} detected. "
                 f"Using {str(NUM_GPUS) + ' GPU(s)' if CUDA_AVAILABLE else 'CPU(Much slower)'}.")
            self.args['gpus'] = list(range(NUM_GPUS))

        if self.args.get('world_size') and self.args.get('dist_backend') == 'gloo':
            """Reserved gloo and world_size for CPU multi process use case."""
            self.args['gpus'] = [None] * self.args.get('world_size')

    def _ddp_setup(self):
        if all([self.args['use_ddp'], len(self.args['gpus']) >= 1]):
            self.args['num_gpus'] = len(self.args['gpus'])
            _os.environ['MASTER_ADDR'] = self.args.get('master_addr', '127.0.0.1')  #
            _os.environ['MASTER_PORT'] = self.args.get('master_port', '12355')
        else:
            self.args['use_ddp'] = False

        """Check if want to do distributed validation"""
        self.args['distributed_validation'] = self.args['use_ddp'] and self.args.get('distributed_validation', False)

    def _show_args(self):
        info('Starting with the following parameters:', self.args['verbose'])
        if self.args['verbose']:
            _pp.pprint(self.args)

    def _init_args(self, args):
        if isinstance(args, _AP):
            self.args = vars(args.parse_args())
        elif isinstance(args, dict):
            self.args = {**args}
        else:
            raise ValueError('2nd Argument of EasyTorch could be only one of :ArgumentParser, dict')

    def _make_reproducible(self):
        if self.args['use_ddp'] and self.args['seed'] is None:
            raise ValueError('Seed must be explicitly given as seed=<seed> (Eg.1, 2, 101, 102) in DDP.')

        if self.args['seed'] is None:
            self.args['seed'] = CURRENT_SEED

        if self.args.get('seed_all'):
            _torch.manual_seed(self.args['seed'])
            _torch.cuda.manual_seed_all(self.args['seed'])
            _torch.cuda.manual_seed(self.args['seed'])
            _np.random.seed(self.args['seed'])
            _random.seed(self.args['seed'])
            _torch.backends.cudnn.deterministic = True
            _torch.backends.cudnn.benchmark = False

    def _init_dataspecs(self, dataspecs):
        """
        Need to add -data(base folder for dataset) to all the directories in dataspecs.
        THis makes it flexible to access dataset from arbitrary location.
        """
        if dataspecs is None or len(dataspecs) == 0:
            dataspecs = [{'name': 'experiment'}]

        self.dataspecs = [{**dspec} for dspec in dataspecs]
        for dspec in self.dataspecs:
            if dspec.get('name') is None:
                raise ValueError('Each dataspecs must have a unique name.')

    def _maybe_advance_run(self, cache):
        r"""
        Checks if there already is a previous run and prompt[Y/N] so that
        we avoid accidentally overriding previous runs and lose temper.
        User can supply -f True flag to override by force.
        """
        if self.args['force']:
            warn('Forced overriding previous logs.', self.args['verbose'])
            return

        nxt = -1
        if _os.path.isdir(self.args['log_dir']):
            for s in _os.listdir(self.args['log_dir']):
                num = s.split('_')[-1]
                try:
                    num = ''.join([d for d in num if d.isdigit()])
                    if int(num) > nxt:
                        nxt = int(num)
                except:
                    pass
        cache['log_dir'] += f"_v{nxt + 1}"

    @staticmethod
    def _init_fold_cache(split_file, cache):
        """ Experiment id is split file name. For the example of k-fold. """
        """ Clear cache to save scores for each fold """

        cache[LogKey.TRAIN_LOG] = []
        cache[LogKey.VALIDATION_LOG] = []
        cache[LogKey.TEST_METRICS] = []

        cache['experiment_id'] = split_file.split('.')[0]
        cache['best_checkpoint'] = f"best_{cache['experiment_id']}{CHK_EXT}"
        cache['latest_checkpoint'] = f"last_{cache['experiment_id']}{CHK_EXT}"
        cache.update(best_val_epoch=0, best_val_score=0.0)
        if cache['metric_direction'] == 'minimize':
            cache['best_val_score'] = MAX_SIZE

    def _global_experiment_end(self, trainer, meter):
        """ Finally, save the global score to a file  """
        trainer.cache[LogKey.GLOBAL_TEST_METRICS].append(
            ['Global', *meter.get()])
        _utils.save_scores(trainer.cache, file_keys=[LogKey.GLOBAL_TEST_METRICS])

        with open(trainer.cache['log_dir'] + _sep + LogKey.SERIALIZABLE_GLOBAL_TEST + '.json', 'w') as f:
            log = {
                'averages': meter.averages.serialize(),
                'metrics': {}
            }

            for mk in meter.metrics:
                log['metrics'][mk] = meter.metrics[mk].serialize()

            f.write(_json.dumps(log))

    def _train(self, trainer, train_dataset, validation_dataset):
        trainer.train(train_dataset, validation_dataset)
        trainer.save_checkpoint(trainer.cache['log_dir'] + _sep + trainer.cache['latest_checkpoint'])
        _utils.save_cache({**self.args, **trainer.cache, **trainer.data_handle.dataspec},
                          experiment_id=trainer.cache['experiment_id'])
        trainer.cache['_saved'] = True

    def _test(self, split_file, trainer, test_dataset) -> dict:
        best_exists = _os.path.exists(trainer.cache['log_dir'] + _sep + trainer.cache['best_checkpoint'])
        if best_exists and (self.args['phase'] == Phase.TRAIN or self.args['pretrained_path'] is None):
            """ Best model will be split_name.pt in training phase, and if no pretrained path is supplied. """
            trainer.load_checkpoint(trainer.cache['log_dir'] + _sep + trainer.cache['best_checkpoint'],
                                    map_location=trainer.device['gpu'], load_optimizer_state=False)

        """ Run and save experiment test scores """
        test_out = trainer.inference(mode='test', save_predictions=True, datasets=test_dataset)
        test_meter = trainer.reduce_scores([test_out], distributed=False)
        trainer.cache[LogKey.TEST_METRICS] = [[split_file, *test_meter.get()]]
        _utils.save_scores(trainer.cache, experiment_id=trainer.cache['experiment_id'],
                           file_keys=[LogKey.TEST_METRICS])

        if not trainer.cache.get('_saved'):
            _utils.save_cache({**self.args, **trainer.cache, **trainer.data_handle.dataspec},
                              experiment_id=f"{trainer.cache['experiment_id']}_test")

        return test_out

    def run(self, trainer_cls: typing.Type[ETTrainer],
            dataset_cls: typing.Type[ETDataset] = None,
            data_handle_cls: typing.Type[ETDataHandle] = ETDataHandle):
        if self.args.get('use_ddp'):
            _mp.spawn(_ddp_worker, nprocs=self.args['num_gpus'],
                      args=(self, trainer_cls, dataset_cls, data_handle_cls, False))
        else:
            self._run(trainer_cls, dataset_cls, data_handle_cls)

    def _run(self, trainer_cls, dataset_cls, data_handle_cls):
        r"""Run for individual datasets"""
        for dspec in self.dataspecs:

            data_handle = data_handle_cls(args=self.args, dataloader_args=self.dataloader_args, dataspec=dspec)
            trainer = trainer_cls(args=self.args, data_handle=data_handle)
            trainer.init_nn(init_models=False, init_weights=False, init_optimizer=False)

            trainer.cache['log_dir'] = self.args['log_dir'] + _sep + dspec['name']
            if self.args['is_master']:
                self._maybe_advance_run(trainer.cache)
            if self.args['verbose']:
                self._show_args()

            trainer.data_handle.init_dataspec_()
            trainer.data_handle.create_splits(out_dir=trainer.cache.get('log_dir'))

            trainer.cache[LogKey.GLOBAL_TEST_METRICS] = []
            trainer.cache['log_header'] = 'Loss|Accuracy'
            trainer.cache.update(monitor_metric='time', metric_direction='maximize')

            """ Init and Run for each splits. """
            test_accum = []
            trainer.init_experiment_cache()

            _os.makedirs(trainer.cache['log_dir'], exist_ok=True)
            if dspec['split_dir'].endswith('.json') or dspec['split_dir'].endswith('.txt'):
                split = _Path(dspec['split_dir'])
                splits = [str(split.name)]
                data_handle.dataspec['split_dir'] = str(split.parent)
            else:
                splits = sorted(_os.listdir(dspec['split_dir']))

            for split_file in splits:
                self._init_fold_cache(split_file, trainer.cache)
                trainer.init_nn()
                if self.args['phase'] == Phase.TRAIN:
                    train_dataset = trainer.data_handle.get_train_dataset(split_file, dataset_cls=dataset_cls)
                    validation_dataset = trainer.data_handle.get_validation_dataset(
                        split_file, dataset_cls=dataset_cls)
                    self._train(trainer, train_dataset, validation_dataset)

                if self.args['is_master']:
                    test_dataset = trainer.data_handle.get_test_dataset(split_file, dataset_cls=dataset_cls)
                    if test_dataset is not None:
                        test_accum.append(self._test(split_file, trainer, test_dataset))

                if trainer.args.get('use_ddp'):
                    _dist.barrier()

            if self.args['is_master']:
                global_scores = trainer.reduce_scores(test_accum, distributed=False)
                self._global_experiment_end(trainer, global_scores)

            if trainer.args.get('use_ddp'):
                _dist.barrier()
            _cleanup(trainer)

    def run_pooled(self, trainer_cls: typing.Type[ETTrainer],
                   dataset_cls: typing.Type[ETDataset] = None,
                   data_handle_cls: typing.Type[ETDataHandle] = ETDataHandle):
        self.args['pooled_run'] = True
        if self.args.get('use_ddp'):
            _mp.spawn(_ddp_worker, nprocs=self.args['num_gpus'],
                      args=(self, trainer_cls, dataset_cls, data_handle_cls, True))
        else:
            self._run_pooled(trainer_cls, dataset_cls, data_handle_cls)

    def _run_pooled(self, trainer_cls, dataset_cls, data_handle_cls):
        r"""  Run in pooled fashion. """
        data_handle = data_handle_cls(args=self.args, dataloader_args=self.dataloader_args)
        trainer = trainer_cls(args=self.args, data_handle=data_handle)
        trainer.init_nn(init_models=False, init_weights=False, init_optimizer=False)

        trainer.cache['log_dir'] = self.args['log_dir'] + _sep + f'Pooled_{len(self.dataspecs)}'
        if self.args['is_master']:
            self._maybe_advance_run(trainer.cache)
        if self.args['verbose']:
            self._show_args()

        for dspec in self.dataspecs:
            trainer.data_handle.dataspec = dspec
            trainer.data_handle.init_dataspec_()
            trainer.data_handle.create_splits(out_dir=trainer.cache['log_dir'] + _sep + dspec['name'])

        trainer.data_handle.dataspec = {}
        for dspec in self.dataspecs:
            trainer.data_handle.dataspec[dspec['name']] = dspec

        trainer.cache[LogKey.GLOBAL_TEST_METRICS] = []
        trainer.cache['log_header'] = 'Loss|Accuracy'
        trainer.cache.update(monitor_metric='time', metric_direction='maximize')

        """Initialize experiment essentials"""
        test_accum = []
        trainer.init_experiment_cache()
        _os.makedirs(trainer.cache['log_dir'], exist_ok=True)

        split_groups = []
        for dspec in self.dataspecs:
            split_groups.append(sorted(_os.listdir(dspec['split_dir'])))

        for i, splits in enumerate(zip(*split_groups)):
            self._init_fold_cache(f'fold_{i}.none', trainer.cache)

            trainer.init_nn()
            if self.args['phase'] == Phase.TRAIN:
                train_dataset = _mproc.pooled_load(
                    'train',
                    self.dataspecs,
                    splits,
                    args=self.args,
                    dataset_cls=dataset_cls,
                    diskcache=data_handle.diskcache,
                    load_sparse=False
                )[0]

                val_dataset = _mproc.pooled_load(
                    'validation',
                    self.dataspecs,
                    splits,
                    args=self.args,
                    dataset_cls=dataset_cls,
                    diskcache=data_handle.diskcache,
                    load_sparse=False
                )[0]

                self._train(trainer, train_dataset, val_dataset)

            """Only do test in master rank node"""
            if self.args['is_master']:
                test_dataset = _mproc.pooled_load(
                    'test',
                    self.dataspecs,
                    splits,
                    args=self.args,
                    dataset_cls=dataset_cls,
                    diskcache=data_handle.diskcache,
                    load_sparse=self.args['load_sparse']
                )
                test_accum.append(self._test('Pooled', trainer, test_dataset))

            if trainer.args.get('use_ddp'):
                _dist.barrier()

        if self.args['is_master']:
            global_scores = trainer.reduce_scores(test_accum, distributed=False)
            self._global_experiment_end(trainer, global_scores)

        if trainer.args.get('use_ddp'):
            _dist.barrier()
        _cleanup(trainer)



import numpy as _np
import cv2 as _cv2
import random as _random
from PIL import Image as _IMG


class RandomGaussJitter:
    def __init__(self, gauss_factor=0.2, jitter_factor=0.3):
        self.gauss_factor = gauss_factor
        self.jitter_factor = jitter_factor
        self.gauss_factor = _np.arange(0, self.gauss_factor + 0.01, 0.05)
        self.jitter_factor = _np.arange(0, self.jitter_factor + 0.01, 0.05)

    def __call__(self, img):
        arr = _np.array(img) / 255.0
        noise = _np.abs(_np.random.rand(*arr.shape)) * _random.choice(self.gauss_factor)
        jitt = _np.random.randint(0, 2, arr.shape)
        noise[jitt == 1] *= _random.choice(self.jitter_factor)
        arr = ((arr + noise) * 255).astype(_np.uint8)

        pick = _random.random()
        if pick > 0.67:
            arr = _cv2.GaussianBlur(arr, (3, 3), 0)
            arr = _cv2.medianBlur(arr, 3)
        elif pick > 0.33:
            arr = _cv2.medianBlur(arr, 3)
            arr = _cv2.GaussianBlur(arr, (3, 3), 0)
        else:
            arr = _cv2.GaussianBlur(arr, (3, 3), 0)

        return _IMG.fromarray(arr)



from easytorch.config import default_ap, default_args
from easytorch.data import ETDataset, ETDataHandle, UnPaddedDDPSampler
from easytorch.metrics import ETMetrics, ETAverages, Prf1a, ConfusionMatrix, ETMeter, AUCROCMetrics

from .easytorch import EasyTorch
from .trainer import ETTrainer



import glob as _glob
import json as _json
import math as _math
import os as _os
from os import sep as _sep

import torch as _torch
import torch.distributed as _dist
import torch.utils.data as _data
from torch.utils.data import DataLoader as _DataLoader, Dataset as _Dataset

import easytorch.data.datautils as _du
import easytorch.utils as _etutils
import easytorch.data.multiproc as _multi
from easytorch.utils.logger import *
import pickle as _pickle
import shutil as _shu
import uuid as _uuid


class DiskCache:
    def __init__(self, path, verbose=True):
        self.path = path
        self.verbose = verbose

    def _gen_key(self, name):
        return _os.path.basename(name) + '-' + _uuid.uuid5(_uuid.NAMESPACE_X500, f"{name}").hex[:8]

    def add(self, name, value):
        _os.makedirs(self.path, exist_ok=True)
        key = self._gen_key(name)
        with open(self.path + _os.sep + key + ".pkl", 'wb') as file:
            _pickle.dump(value, file, _pickle.HIGHEST_PROTOCOL)
        return key

    def get(self, key):
        with open(self.path + _os.sep + key + ".pkl", 'rb') as file:
            return _pickle.load(file)

    def clear(self):
        if _os.path.exists(self.path):
            _shu.rmtree(self.path, ignore_errors=True)
            info(f"Diskcache : {self.path} cleared.", self.verbose)


class ETDataHandle:

    def __init__(self, args=None, dataloader_args=None, dataspec=None, **kw):
        self.args = _etutils.FrozenDict(args)
        self.dataloader_args = _etutils.FrozenDict(dataloader_args)
        self.datasets = {}
        self.dataspec = dataspec
        self.diskcache = DiskCache(
            self.args['log_dir'] + _os.sep + "_cache" + _sep + self.args['RUN-ID'],
            self.args['verbose']
        )

    def get_dataset(self, handle_key, files, reuse=True, dataset_cls=None):
        if reuse and self.datasets.get(handle_key):
            return self.datasets[handle_key]
        dataset = dataset_cls(mode=handle_key, limit=self.args['load_limit'], **self.args)
        dataset.add(files=files, diskcache=self.diskcache, verbose=self.args['verbose'], **self.dataspec)
        if reuse:
            self.datasets[handle_key] = dataset
        return dataset

    def get_train_dataset(self, split_file, dataset_cls=None):
        if dataset_cls is None or self.dataloader_args.get('train', {}).get('dataset'):
            return self.dataloader_args.get('train', {}).get('dataset')

        r"""Load the train data from current fold/split."""
        with open(self.dataspec['split_dir'] + _sep + split_file) as file:
            split = _json.loads(file.read())
            train_dataset = self.get_dataset('train', split.get('train', []), dataset_cls=dataset_cls)
            return train_dataset

    def get_validation_dataset(self, split_file, dataset_cls=None):
        if dataset_cls is None or self.dataloader_args.get('validation', {}).get('dataset'):
            return self.dataloader_args.get('validation', {}).get('dataset')

        r""" Load the validation data from current fold/split."""
        with open(self.dataspec['split_dir'] + _sep + split_file) as file:
            split = _json.loads(file.read())
            val_dataset = self.get_dataset('validation', split.get('validation', []), dataset_cls=dataset_cls)
            if val_dataset and len(val_dataset) > 0:
                return val_dataset

    def get_test_dataset(self, split_file, dataset_cls=None):
        if dataset_cls is None or self.dataloader_args.get('test', {}).get('dataset'):
            return self.dataloader_args.get('test', {}).get('dataset')

        with open(self.dataspec['split_dir'] + _sep + split_file) as file:
            if split_file.endswith('.json'):
                _files = _json.loads(file.read()).get('test', [])

            elif split_file.endswith('.txt'):
                _files = file.read().splitlines()

            if self.args['load_sparse']:
                datasets = _multi.multi_load('test', _files, self.dataspec, self.args, dataset_cls, self.diskcache)
                success(f'\n{len(datasets)} sparse dataset loaded.', self.args['verbose'])
            else:
                datasets = self.get_dataset('test', _files, self.dataspec, dataset_cls=dataset_cls)

            if len(datasets) > 0 and sum([len(t) for t in datasets if t]) > 0:
                return datasets

    def get_loader(self, handle_key='', distributed=False, use_unpadded_sampler=False, **kw):
        args = {**self.args}
        args['distributed'] = distributed
        args['use_unpadded_sampler'] = use_unpadded_sampler
        args.update(self.dataloader_args.get(handle_key, {}))
        args.update(**kw)

        if args.get('dataset') is None:
            return None

        loader_args = {
            'dataset': None,
            'batch_size': 1,
            'sampler': None,
            'shuffle': False,
            'batch_sampler': None,
            'num_workers': 0,
            'pin_memory': False,
            'drop_last': False,
            'timeout': 0,
            'worker_init_fn': _multi.seed_worker if args.get('seed_all') else None
        }

        for k in loader_args.keys():
            loader_args[k] = args.get(k, loader_args.get(k))

        if loader_args['sampler'] is not None:
            """Sampler and shuffle are mutually exclusive"""
            loader_args['shuffle'] = False

        if args['distributed']:
            sampler_args = {
                'num_replicas': args.get('replicas'),
                'rank': args.get('rank'),
                'shuffle': args.get('shuffle'),
                'seed': args.get('seed')
            }

            if loader_args.get('sampler') is None:
                loader_args['shuffle'] = False  # Shuffle is mutually exclusive with sampler
                if args['use_unpadded_sampler']:
                    loader_args['sampler'] = UnPaddedDDPSampler(loader_args['dataset'], **sampler_args)
                else:
                    loader_args['sampler'] = _data.distributed.DistributedSampler(loader_args['dataset'],
                                                                                  **sampler_args)

            loader_args['num_workers'] = _multi.num_workers(args, loader_args, True)
            loader_args['batch_size'] = _multi.batch_size(args, loader_args, True)

        return _DataLoader(collate_fn=_multi.safe_collate, **loader_args)

    def create_splits(self, out_dir):
        if _du.should_create_splits_(out_dir, self.dataspec, self.args):
            _du.default_data_splitter_(files=self._list_files(), dspec=self.dataspec, args=self.args)
            info(f"{len(_os.listdir(self.dataspec['split_dir']))} split(s) created in "
                 f"'{self.dataspec['split_dir']}' directory.", self.args['verbose'])
        else:
            splits_len = len(_os.listdir(self.dataspec['split_dir']))
            info(f"{splits_len} split(s) loaded from '{self.dataspec['split_dir']}' directory.",
                 self.args['verbose'] and splits_len > 0)

    def _list_files(self) -> list:
        ext = self.dataspec.get('extension', '*').replace('.', '')
        rec = self.dataspec.get('recursive', False)
        rec_pattern = '**/' if rec else ''
        if self.dataspec.get('sub_folders') is None:
            _path = self.dataspec['data_dir']
            _pattern = f"{_path}/{rec_pattern}*.{ext}"
            _files = _glob.glob(_pattern, recursive=rec)
            return [f.replace(_path + _sep, '') for f in _files]

        files = []
        for sub in self.dataspec['sub_folders']:
            path = self.dataspec['data_dir'] + _sep + sub
            files += [f.replace(self.dataspec['data_dir'] + _sep, '') for f in
                      _glob.glob(f"{path}/{rec_pattern}*.{ext}", recursive=rec)]
        return files

    def init_dataspec_(self):
        for k in self.dataspec:
            if '_dir' in k:
                path = _os.path.join(self.args['dataset_dir'], self.dataspec[k])
                path = path.replace(f"{_sep}{_sep}", _sep)
                if path.endswith(_sep):
                    path = path[:-1]
                self.dataspec[k] = path


class KFoldDataHandle(ETDataHandle):
    """Use this when needed to run k-fold(train,test one each fold) on directly passed Dataset from dataloader_args"""

    def create_splits(self, out_dir):
        if self.args.get('num_folds') is None:
            super(KFoldDataHandle, self).create_splits(out_dir)
        else:
            self.dataspec['split_dir'] = out_dir + _os.sep + 'splits'
            _os.makedirs(self.dataspec['split_dir'], exist_ok=True)
            _du.create_k_fold_splits(
                list(range(len(self.dataloader_args['train']['dataset']))), self.args['num_folds'],
                save_to_dir=self.dataspec['split_dir']
            )

    def get_test_dataset(self, split_file, dataset_cls=None):
        if self.args.get('num_folds') is None:
            return super(KFoldDataHandle, self).get_test_dataset(split_file, dataset_cls)
        else:
            with open(self.dataspec['split_dir'] + _os.sep + split_file) as file:
                test_ix = _json.loads(file.read()).get('test', [])
                return _data.Subset(self.dataloader_args['train']['dataset'], test_ix)

    def get_train_dataset(self, split_file, dataset_cls=None):
        if self.args.get('num_folds') is None:
            return super(KFoldDataHandle, self).get_train_dataset(split_file, dataset_cls)
        else:
            with open(self.dataspec['split_dir'] + _os.sep + split_file) as file:
                train_ix = _json.loads(file.read()).get('train', [])
                return _data.Subset(self.dataloader_args['train']['dataset'], train_ix)

    def get_validation_dataset(self, split_file, dataset_cls=None):
        if self.args.get('num_folds') is None:
            return super(KFoldDataHandle, self).get_validation_dataset(split_file, dataset_cls)
        else:
            with open(self.dataspec['split_dir'] + _os.sep + split_file) as file:
                val_ix = _json.loads(file.read()).get('validation', [])
                return _data.Subset(self.dataloader_args['train']['dataset'], val_ix)


class ETDataset(_Dataset):
    def __init__(self, mode='init', limit=None, **kw):
        self.mode = mode
        self.limit = limit
        self.indices = []
        self.data = {}
        self.diskcache = None
        self.args = _etutils.FrozenDict(kw)
        self.dataspecs = _etutils.FrozenDict({})

    def load_index(self, dataset_name, file):
        r"""
        Logic to load indices of a single file.
        -Sometimes one image can have multiple indices like U-net where we have to get multiple patches of images.
        """
        self.indices.append([dataset_name, file])

    def _load_indices(self, dataspec_name, files, verbose=True):
        r"""
        We load the proper indices/names(whatever is called) of the files in order to prepare minibatches.
        Only load lim numbr of files so that it is easer to debug(Default is infinite, -lim/--load-lim argument).
        """
        _files = files[:self.limit]
        _files_len = len(_files)
        if self.args["multi_load"] and _files_len > 1:
            dataset_objs = _multi.multi_load(
                self.mode,
                _files,
                self.dataspecs[dataspec_name],
                self.args,
                self.__class__,
                diskcache=self.diskcache
            )
            self.gather(dataset_objs)
        else:
            for i, file in enumerate(_files, 1):
                info(f"Loading... {i}/{_files_len}", i % _multi.LOG_FREQ == 0)
                self.load_index(dataspec_name, file)
        success(f'{dataspec_name}, {self.mode}, {len(self)} indices Loaded.', verbose)

    def gather(self, dataset_objs):
        for d in dataset_objs:
            attributes = vars(d)
            for k, v in attributes.items():
                if isinstance(v, _etutils.FrozenDict):
                    continue

                if isinstance(v, list):
                    self.__getattribute__(f"{k}").extend(v)

                elif isinstance(attributes[f"{k}"], dict):
                    self.__getattribute__(f"{k}").update(**v)

                elif isinstance(attributes[f"{k}"], set):
                    self.__getattribute__(f"{k}").union(v)

    def __getitem__(self, index):
        r"""
        Logic to load one file and send to model. The mini-batch generation will be handled by Dataloader.
        Here we just need to write logic to deal with single file.
        """
        raise NotImplementedError('Must be implemented by child class.')

    def __len__(self):
        return len(self.indices)

    def add(self, files, diskcache=None, verbose=True, **kw):
        r""" An extra layer for added flexibility."""
        self.diskcache = diskcache
        self.dataspecs[kw['name']] = kw
        self._load_indices(dataspec_name=kw['name'], files=files, verbose=verbose)


class UnPaddedDDPSampler(_data.Sampler):
    r"""fork from official pytorch repo: torch.data.distributed.DistributedSampler where padding is off"""
    r"""https://github.com/pytorch/"""

    r"""Sampler that restricts data loading to a subset of the dataset.

    It is especially useful in conjunction with
    :class:`torch.nn.parallel.DistributedDataParallel`. In such a case, each
    process can pass a :class`~torch.utils.data.DistributedSampler` instance as a
    :class:`~torch.utils.data.DataLoader` sampler, and load a subset of the
    original dataset that is exclusive to it.

    .. note::
        Dataset is assumed to be of constant size.

    Arguments:
        dataset: Dataset used for sampling.
        num_replicas (int, optional): Number of processes participating in
            distributed training. By default, :attr:`rank` is retrieved from the
            current distributed group.
        rank (int, optional): Rank of the current process within :attr:`num_replicas`.
            By default, :attr:`rank` is retrieved from the current distributed
            group.
        shuffle (bool, optional): If ``True`` (default), sampler will shuffle the
            indices.
        seed (int, optional): random seed used to shuffle the sampler if
            :attr:`shuffle=True`. This number should be identical across all
            processes in the distributed group. Default: ``0``.

    .. warning::
        In distributed mode, calling the :meth`set_epoch(epoch) <set_epoch>` method at
        the beginning of each epoch **before** creating the :class:`DataLoader` iterator
        is necessary to make shuffling work properly across multiple epochs. Otherwise,
        the same ordering will be always used.

    Example::

        >>> sampler = DistributedSampler(dataset) if is_distributed else None
        >>> loader = DataLoader(dataset, shuffle=(sampler is None),
        ...                     sampler=sampler)
        >>> for epoch in range(start_epoch, n_epochs):
        ...     if is_distributed:
        ...         sampler.set_epoch(epoch)
        ...     train(loader)
    """

    def __init__(self, dataset, num_replicas=None, rank=None, shuffle=True, seed=0):
        if num_replicas is None:
            if not _dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            num_replicas = _dist.get_world_size()
        if rank is None:
            if not _dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            rank = _dist.get_rank()
        self.dataset = dataset
        self.num_replicas = num_replicas
        self.rank = rank
        self.epoch = 0

        """For unpadded sampling"""
        self.num_samples = int(_math.ceil((len(self.dataset) - self.rank) * 1.0 / self.num_replicas))
        self.total_size = len(self.dataset)

        self.shuffle = shuffle
        self.seed = seed

    def __iter__(self):
        if self.shuffle:
            # deterministically shuffle based on epoch and seed
            g = _torch.Generator()
            g.manual_seed(self.seed + self.epoch)
            indices = _torch.randperm(len(self.dataset), generator=g).tolist()
        else:
            indices = list(range(len(self.dataset)))

        """Do not pad anything"""
        # add extra samples to make it evenly divisible
        # indices += indices[:(self.total_size - len(indices))]

        assert len(indices) == self.total_size

        # subsample
        indices = indices[self.rank:self.total_size:self.num_replicas]
        assert len(indices) == self.num_samples

        return iter(indices)

    def __len__(self):
        return self.num_samples

    def set_epoch(self, epoch):
        r"""
        Sets the epoch for this sampler. When :attr:`shuffle=True`, this ensures all replicas
        use a different random ordering for each epoch. Otherwise, the next iteration of this
        sampler will yield the same ordering.

        Arguments:
            epoch (int): Epoch number.
        """
        self.epoch = epoch



import random as _random
import sys as _sys
from enum import Enum as _Enum

import torch as _torch

MYSELF = 'easytorch'

CUDA_AVAILABLE = _torch.cuda.is_available()
NUM_GPUS = _torch.cuda.device_count()

METRICS_EPS = 1e-5
METRICS_NUM_PRECISION = 5

MAX_SIZE = _sys.maxsize
DATA_SPLIT_RATIO = [0.6, 0.2, 0.2]
CURRENT_SEED = _random.randint(0, 2 ** 24)
SCORE_DELTA = 0.0

SCORE_MIN = 0.0
SCORE_MAX = 1.0

CHK_EXT = '.tar'

MASTER_RANK = 0


class Phase(str, _Enum):
    TRAIN = 'train'
    VALIDATION = 'validation'
    TEST = 'test'
    EVAL = 'eval'


class LogKey(str, _Enum):
    TRAIN_LOG = 'train_log'
    SERIALIZABLE_TRAIN = 'serializable_train_scores'

    VALIDATION_LOG = 'validation_log'
    VALIDATION_METRICS = "validation_metrics"
    SERIALIZABLE_VALIDATION = 'serializable_validation_scores'

    TEST_METRICS = 'test_metrics'
    SERIALIZABLE_TEST = 'serializable_test_scores'

    GLOBAL_TEST_LOG = 'global_test_log'
    GLOBAL_TEST_METRICS = 'global_test_metrics'
    SERIALIZABLE_GLOBAL_TEST = 'serializable_global_test_scores'



import os as _os
import random as _random
import warnings as _warn
from abc import ABC as _ABC

import numpy as _np
import torchvision.transforms as _tmf
from easytorch.data import ETDataset as _ETDataset
import easytorch.vision.imageutils as _imgutils

sep = _os.sep


class BaseImageDataset(_ETDataset, _ABC):
    def __init__(self, **kw):
        super().__init__(**kw)
        self.transforms = self.get_transforms()
        self.pil_to_tensor = _tmf.Compose([_tmf.ToPILImage(), _tmf.ToTensor()])

    def get_transforms(self):
        return _tmf.Compose([_tmf.ToPILImage(), _tmf.ToTensor()])

    def _validate_image_data(self, dspec, img_obj):
        return img_obj

    def load_img(self, dspec, file):
        img_obj = _imgutils.Image()

        """Load Image"""
        img_obj.load(dspec['data_dir'], file)
        img_obj.apply_clahe()
        if len(img_obj.array.shape) > 2 and img_obj.array.shape[-1] != 3:
            _warn.warn(f"Suspicious Image shape: {img_obj.array.shape}, clipping to RGB: {dspec['name']}")
            img_obj.array = img_obj.array[:, :, :3]

        """Load ground truth"""
        dspec['has_gt'] = any(['label_dir' in dspec.keys()])
        if dspec['has_gt']:
            img_obj.load_ground_truth(dspec["label_dir"], dspec["label_getter"])
        else:
            _warn.warn(f"Random label initialized: {dspec['name']}")
            img_obj.ground_truth = _np.random.randint(0, 2, img_obj.array.shape[:2]).astype(_np.uint8) * 255

        """Load mask"""
        dspec['has_mask'] = any(['mask_dir' in dspec.keys()])
        if dspec['has_mask']:
            img_obj.load_mask(dspec['mask_dir'], dspec['mask_getter'])

        return self._validate_image_data(dspec, img_obj)

    def __getitem__(self, index):
        raise NotImplementedError('Must implement')


class PatchedImgDataset(BaseImageDataset, _ABC):
    """dataspec must have patch_shape, patch_offset, """

    def load_index(self, dataset_name, file):
        r"""
        :param dataset_name: name of teh dataset as provided in train_dataspecs
        :param file: Name of an image
        :return:
        Logic split an image to patches and feed to U-Net. Meanwhile we need to store the four-corners
            of each patch so that we can rejoin the full image from the patches' corresponding predictions.
        """
        dt = self.dataspecs[dataset_name]
        obj = self.load_img(dt, file)

        cache_key = self.diskcache.add(f"{dataset_name}_{file}", obj)
        for corners in _imgutils.get_chunk_indexes(
                obj.array.shape[:2],
                dt['patch_shape'],
                dt['patch_offset']
        ):
            """
            get_chunk_indexes will return the list of four corners of all patches of the images  
            by using window size of self.patch_shape, and offset  of elf.patch_offset
            """
            self.indices.append([dataset_name, file] + corners + [cache_key])


class BinaryPatchDataset(PatchedImgDataset):
    def __init__(self, **kw):
        super().__init__(**kw)

    def _validate_image_data(self, dspec, img_obj):
        thr_manual = dspec.setdefault('thr_manual', 50)
        if dspec.get('has_gt'):
            gt_unique = _np.unique(img_obj.ground_truth)
            if len(img_obj.ground_truth.shape) > 2:
                _warn.warn(
                    f"Ground truth shape suspicious: {img_obj.ground_truth.shape}, "
                    f"using 1st channel onl:  {dspec['name']}"
                )
                img_obj.ground_truth = img_obj.ground_truth[:, :, 0]

            if sum(gt_unique) == 1:
                _warn.warn(f"Ground truth 1 converted to 255")
                img_obj.ground_truth[img_obj.ground_truth == 1] = 255

            if len(gt_unique) != 2:
                _warn.warn(
                    f"Number of unique ground truth items != {2} in  {dspec['name']}. "
                    f"\nBinarizing ... {gt_unique}: "
                )
                _imgutils.binarize(img_obj.ground_truth, thr_manual)

        if dspec.get('has_mask'):
            mask_unique = _np.unique(img_obj.mask)
            if len(img_obj.mask.shape) > 2:
                _warn.warn(f"Mask shape suspicious: {img_obj.mask.shape}, using 1st channel only: {dspec['name']}")
                img_obj.mask = img_obj.mask[:, :, 0]

            if sum(mask_unique) == 1:
                _warn.warn(f"Mask truth 1 converted to 255")
                img_obj.mask[img_obj.mask == 1] = 255

            if len(mask_unique) != 2:
                _warn.warn(
                    f"Number of unique items in mask: {mask_unique} in {dspec['name']}"
                )

        if dspec.get('bbox_crop'):
            copy = img_obj.copy()
            img_obj.array, img_obj.ground_truth, img_obj.mask = _imgutils.masked_bboxcrop(
                img_obj.array,
                img_obj.ground_truth,
                offset=dspec.setdefault('bbox_crop_offset', 51),
                threshold=dspec.setdefault('bbox_crop_threshold', 5)
            )

            if img_obj.array.shape[0] < dspec['patch_shape'][0] or img_obj.array.shape[1] < dspec['patch_shape'][1]:
                _warn.warn(
                    f"BBOX crop reversing for "
                    f"{dspec['name']}:{img_obj.file}, shape: {img_obj.array.shape}, {copy.array.shape}"

                )
                img_obj = copy.copy()

        if dspec.get('resize'):
            img_obj.array = _imgutils.resize(img_obj.array, dspec['resize'])
            img_obj.ground_truth = _imgutils.resize(img_obj.ground_truth, dspec['resize'])

            if img_obj.mask is not None:
                img_obj.mask = _imgutils.resize(img_obj.mask, dspec['resize'])

        """Must binarize after resize"""
        if dspec.get('resize'):
            _imgutils.binarize(img_obj.ground_truth, thr_manual)
            if dspec.get('has_mask'):
                _imgutils.binarize(img_obj.mask, thr_manual)
        return img_obj

    def __getitem__(self, index):
        """
        :param index:
        :return: dict with keys - indices, input, label
            We need indices to get the file name to save the respective predictions.
        """
        dname, file, row_from, row_to, col_from, col_to, cache_key = self.indices[index]

        obj = self.diskcache.get(cache_key)
        img = obj.array
        gt = obj.ground_truth[row_from:row_to, col_from:col_to]

        p, q, r, s, pad = _imgutils.expand_and_mirror_patch(
            img.shape,
            [row_from, row_to, col_from, col_to],
            self.dataspecs[dname]['expand_by']
        )
        if len(img.shape) == 3:
            pad = [*pad, (0, 0)]

        img = _np.pad(img[p:q, r:s], pad, 'reflect')
        if self.mode == 'train' and _random.uniform(0, 1) <= 0.5:
            img = _np.flip(img, 0)
            gt = _np.flip(gt, 0)

        if self.mode == 'train' and _random.uniform(0, 1) <= 0.5:
            img = _np.flip(img, 1)
            gt = _np.flip(gt, 1)

        img = self.transforms(img)
        gt = self.pil_to_tensor(gt)
        return {'indices': self.indices[index], 'input': img, 'label': gt.squeeze()}


class FullImgDataset(BaseImageDataset):
    def __init__(self, **kw):
        r"""
        Initialize necessary shapes for unet.
        """
        super().__init__(**kw)
        self.labels = None

    def get_transforms(self):
        if self.mode == "train":
            return _tmf.Compose(
                [_tmf.ToPILImage(),
                 _tmf.RandomHorizontalFlip(),
                 _tmf.RandomVerticalFlip(),
                 _tmf.ToTensor()]
            )
        return self.pil_to_tensor

    def _load_labels(self, dspec):
        return None

    def _get_label(self, file):
        pass

    def _validate_image_data(self, dspec, img_obj):

        if dspec.get('bbox_crop'):
            copy = img_obj.copy()
            img_obj.array, img_obj.mask = _imgutils.masked_bboxcrop(
                img_obj.array,
                offset=dspec.setdefault('bbox_crop_offset', 31),
                threshold=dspec.setdefault('bbox_crop_threshold', 5)
            )

            if dspec.get('resize') and (
                    img_obj.array.shape[0] < 0.55 * copy.array.shape[0]
                    or
                    img_obj.array.shape[1] < 0.55 * copy.array.shape[1]
            ):
                _warn.warn(f"BBOX crop reversing for {dspec['name']}:{img_obj.file}, shape: {img_obj.array.shape}")
                img_obj = copy.copy()

        if dspec.get('resize'):
            img_obj.array = _imgutils.resize(img_obj.array, dspec['resize'])

            if img_obj.mask is not None:
                img_obj.mask = _imgutils.resize(img_obj.mask, dspec['resize'])

        return img_obj

    def load_img(self, dspec, file):
        img_obj = _imgutils.Image()

        """Load Image"""
        img_obj.load(dspec['data_dir'], file)
        img_obj.apply_clahe()
        if len(img_obj.array.shape) > 2 and img_obj.array.shape[-1] != 3:
            _warn.warn(f"Suspicious Image shape: {img_obj.array.shape}, clipping to RGB: {dspec['name']}")
            img_obj.array = img_obj.array[:, :, :3]

        """Load mask"""
        dspec['has_mask'] = any(['mask_dir' in dspec.keys()])
        if dspec['has_mask']:
            img_obj.load_mask(dspec['mask_dir'], dspec['mask_getter'])

        return self._validate_image_data(dspec, img_obj)

    def load_index(self, dataset_name, file):
        dspec = self.dataspecs[dataset_name]
        if self.labels is None:
            self.labels = self._load_labels(dspec)

        img_obj = self.load_img(dspec, file)
        cache_key = self.diskcache.add(f"{dataset_name}_{file}", img_obj)
        self.indices.append([dataset_name, file, self._get_label(file), cache_key])

    def __getitem__(self, index):
        _dname, file, label, cache_key = self.indices[index]
        obj = self.diskcache.get(cache_key)
        img = obj.array

        img = self.transforms(img)
        return {'indices': self.indices[index], 'input': img, 'label': _np.array(label)}



import pathlib
import subprocess

from setuptools import setup

# The directory containing this file
_HERE = pathlib.Path(__file__).parent

# The text of the README file
_README = (_HERE / "README.md").read_text()

# Requirements
_requires = ['numpy',
             'scipy',
             'scikit-learn',
             'scikit-image',
             'pillow',
             'matplotlib',
             'pandas',
             'seaborn']

subprocess.call(["pip", "install", "--upgrade", "pip"])

try:
    import cv2
except:
    _requires.append('opencv-contrib-python-headless')

# This call to setup() does all the work
setup(
    name="easytorch",
    version="3.4.6",
    description="Easy Neural Network Experiments with pytorch",
    long_description=_README,
    long_description_content_type="text/markdown",
    url="https://github.com/sraashis/easytorch",
    author="Aashis Khana1",
    author_email="sraashis@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
    ],
    packages=['easytorch', 'easytorch.config', 'easytorch.data', 'easytorch.metrics', 'easytorch.utils',
              'easytorch.vision'],
    include_package_data=True,
    install_requires=_requires
)



r"""
ETTrainer calls .averages(), .metrics() internally to get whatever you have
added in the ETAverages, ETMetrics respectively.
"""

import abc as _abc
import time as _time
from typing import List

import numpy as _np
import torch
import torch as _torch
import torch.distributed as _dist
from sklearn import metrics as _metrics
import json as _json

from easytorch.config.state import *


class ETMetrics:
    def __init__(self, device='cpu', **kw):
        self.device = device

    def serialize(self, skip_attributes=[]):
        _self = vars(self)
        attributes = {}
        for k in set(_self) - set(skip_attributes):

            if isinstance(_self[k], _np.ndarray):
                attributes[k] = _self[k].tolist()

            elif isinstance(_self[k], _torch.Tensor):
                attributes[k] = _self[k].cpu().tolist()

            elif isinstance(_self[k], ETMetrics) or isinstance(_self[k], ETAverages):
                attributes[k] = _self[k].serialize()

            else:
                try:
                    _json.dumps(_self[k])
                    attributes[k] = _self[k]
                except (TypeError, OverflowError):
                    attributes[k] = f"{_self[k]}"

        return attributes

    @_abc.abstractmethod
    def add(self, *args, **kw):
        r"""
        Add two tensor to collect scores.
        Example implementation easytorch.metrics.Prf1a().
        Calculate/store all True Positives, False Positives, True Negatives, False Negatives:
           out = F.softmax(core(x), 1)
           _, pred = torch.max(out, 1)
           sc = self.new_metrics()
           sc.add(pred, labels)
        """
        raise NotImplementedError('Must be implemented like add(self, pred, labels):{...}')

    @_abc.abstractmethod
    def accumulate(self, other):
        r"""
        Add all the content from another ETMetrics object.
        """
        raise NotImplementedError('Must be implemented to accumulate other to self')

    @_abc.abstractmethod
    def reset(self):
        r"""
        Clear all the content of self.
        """
        raise NotImplementedError('Must be implemented to reset')

    @_abc.abstractmethod
    def get(self, *args, **kw) -> List[float]:
        r"""
        Computes/returns list of scores.
            Example: easytorch.metrics.Prf1a() returns
            Precision, Recall, F1, Accuracy from the collected TP, TN, FP, FN.
        """
        raise NotImplementedError('Must be implemented to return the computed scores like Accuracy, F1...')

    @property
    def eps(self):
        r"""
        Epsilon(default 10e-5) for numerical stability.
        """
        return METRICS_EPS

    @property
    def num_precision(self):
        r"""
        Numerical Precision(default 5) for nice looking numbers.
        """
        return METRICS_NUM_PRECISION

    @property
    def time(self):
        return _time.time()

    def extract(self, field):
        sc = getattr(self, field)
        if callable(sc):
            sc = sc()
        return sc

    @_abc.abstractmethod
    def dist_gather(self, device='cpu'):
        pass


class ETAverages(ETMetrics):
    def __init__(self, num_averages=1, **kw):
        r"""
        This class can keep track of K averages.
        For example, in GAN we need to keep track of Generators loss
        """
        super().__init__(**kw)
        self.values = _np.array([0.0] * num_averages, dtype=_np.float)
        self.counts = _np.array([0.0] * num_averages, dtype=_np.float)
        self.num_averages = num_averages

    def add(self, val=0, n=1, index=0):
        r"""
        Keep adding val, n to get the average later.
        Index is the position on where to add the values.
        For example:
            avg = ETAverages(num_averages=2)
            avg.add(lossG.item(), len(batch), 0)
            avg.add(lossD.item(), len(batch), 1)
        """
        self.values[index] += val * n
        self.counts[index] += n

    def accumulate(self, other):
        r"""
        Add another ETAverage object to self
        """
        self.values += other.values
        self.counts += other.counts

    def reset(self):
        r"""
        Clear all the content of self.
        """
        self.values = _np.array([0.0] * self.num_averages)
        self.counts = _np.array([0.0] * self.num_averages)

    def get(self) -> List[float]:
        r"""
        Computes/Returns self.num_averages number of averages in vectorized way.
        """
        counts = self.counts.copy()
        counts[counts == 0] = _np.inf
        return _np.round(self.values / counts, self.num_precision).tolist()

    def average(self, reduce_mean=True):
        avgs = self.get()
        if reduce_mean:
            return round(sum(avgs) / len(avgs), self.num_precision)
        return avgs

    def dist_gather(self, device='cpu'):
        serial = _torch.from_numpy(_np.array([self.counts, self.values])).to(device)
        _dist.all_reduce(serial, op=_dist.ReduceOp.SUM)
        self.counts, self.values = serial.cpu().numpy()


class ETMeter:
    def __init__(self, num_averages=1, **kw):
        self.averages = ETAverages(num_averages)
        self.metrics = {**kw}

    def get(self):
        res = self.averages.get()
        for mk in self.metrics:
            res += self.metrics[mk].get()
        return res

    def __repr__(self):
        metrics = {}
        for mk in self.metrics:
            metrics[mk] = self.metrics[mk].get()

        avg = self.averages.get()
        if self.metrics:
            return f"{avg}, {metrics}"

        return f"{avg}"

    def extract(self, field):
        for mk in self.metrics:
            try:
                return mk, self.metrics[mk].extract(field)
            except:
                pass

        return 'averages', self.averages.extract(field)

    def reset(self):
        if self.averages:
            self.averages.reset()

        for mk in self.metrics:
            self.metrics[mk].reset()

    def accumulate(self, meter):
        self.averages.accumulate(meter.averages)

        for mk in meter.metrics:
            self.metrics[mk].accumulate(meter.metrics[mk])


class Prf1a(ETMetrics):
    r"""
    A class that has GPU based computation for binary classification of:
        Precision, Recall, F1 Score, Accuracy, and Overlap(IOU).
    """

    def __init__(self, **kw):
        super().__init__(**kw)
        self.tn, self.fp, self.fn, self.tp = 0, 0, 0, 0

    def add(self, pred: _torch.Tensor, true: _torch.Tensor):
        y_true = true.view(-1).int().squeeze()
        y_pred = pred.view(-1).int().squeeze()

        y_true[y_true == 255] = 1
        y_pred[y_pred == 255] = 1

        y_true = y_true * 2
        y_cases = y_true + y_pred
        self.tp += _torch.sum(y_cases == 3).item()
        self.fp += _torch.sum(y_cases == 1).item()
        self.tn += _torch.sum(y_cases == 0).item()
        self.fn += _torch.sum(y_cases == 2).item()

    def accumulate(self, other: ETMetrics):
        self.tp += other.tp
        self.fp += other.fp
        self.tn += other.tn
        self.fn += other.fn

    def reset(self):
        self.tn, self.fp, self.fn, self.tp = [0] * 4

    @property
    def precision(self):
        p = self.tp / max(self.tp + self.fp, self.eps)
        return round(p, self.num_precision)

    @property
    def recall(self):
        r = self.tp / max(self.tp + self.fn, self.eps)
        return round(r, self.num_precision)

    @property
    def accuracy(self):
        a = (self.tp + self.tn) / \
            max(self.tp + self.fp + self.fn + self.tn, self.eps)
        return round(a, self.num_precision)

    @property
    def f1(self):
        return self.f_beta(beta=1)

    def f_beta(self, beta=1):
        f_beta = (1 + beta ** 2) * self.precision * self.recall / \
                 max(((beta ** 2) * self.precision) + self.recall, self.eps)
        return round(f_beta, self.num_precision)

    def get(self):
        return [self.accuracy, self.f1, self.precision, self.recall]

    @property
    def overlap(self):
        o = self.tp / max(self.tp + self.fp + self.fn, self.eps)
        return round(o, self.num_precision)

    def dist_gather(self, device='cpu'):
        serial = _torch.from_numpy(_np.array([self.tn, self.fp, self.fn, self.tp])).to(device)
        _dist.all_reduce(serial, op=_dist.ReduceOp.SUM)
        self.tn, self.fp, self.fn, self.tp = serial.cpu().numpy().tolist()


class AUCROCMetrics(ETMetrics):
    __doc__ = "Restricted to binary case"

    def __init__(self, **kw):
        super().__init__(**kw)
        self.probabilities = []
        self.labels = []
        self.thresholds = None
        self._auc = 0

    def accumulate(self, other):
        self.probabilities += other.probabilities
        self.labels += other.labels

    def reset(self):
        self.probabilities = []
        self.labels = []

    def auc(self):
        if len(self.labels) > 0:
            fpr, tpr, self.thresholds = _metrics.roc_curve(self.labels, self.probabilities, pos_label=1)
            return max(_metrics.auc(fpr, tpr), self._auc)
        return 0.0

    def get(self, *args, **kw) -> List[float]:
        return [round(self.auc(), self.num_precision)]

    def dist_gather(self, device='cpu'):
        auc = _torch.from_numpy(_np.array([self.auc()])).to(device)
        _dist.all_reduce(auc, op=_dist.ReduceOp.SUM)
        self._auc = float(auc.item() / _dist.get_world_size())

    def add(self, pred: _torch.Tensor, true: _torch.Tensor):
        self.probabilities += pred.flatten().clone().detach().cpu().tolist()
        self.labels += true.clone().flatten().detach().cpu().tolist()

    def serialize(self, skip_attributes=[]):
        return super(AUCROCMetrics, self).serialize(skip_attributes=['probabilities', 'labels'])


class ConfusionMatrix(ETMetrics):
    """
    Confusion matrix  is used in multi class classification case.
    x-axis is predicted. y-axis is true label.(Like sklearn)
    F1 score from average precision and recall is calculated
    multilabel: N * 2 * C
    """

    def __init__(self, num_classes=None, multilabel=False, **kw):
        super().__init__(**kw)
        self.multilabel = multilabel
        self.num_classes = num_classes
        self.matrix_eps = None
        self.matrix = None
        self.prfa = None
        self.reset()

    def reset(self):
        if self.multilabel:
            self.matrix = _torch.zeros(self.num_classes, 2, 2, device=self.device).long()
            self.prfa = Prf1a()
        else:
            self.matrix = _torch.zeros(self.num_classes, self.num_classes, device=self.device).long()
            self.matrix_eps = torch.from_numpy(_np.array([self.eps] * self.num_classes)).to(self.device)

    def accumulate(self, other: ETMetrics):
        self.matrix += other.matrix

    def add(self, pred: _torch.Tensor, true: _torch.Tensor):
        """
        :param pred: N * 2 * ...
        :param true: N * 2 * ...
        Computes macro F1 by Default.
        """

        if self.multilabel and len(pred.shape) == 2 and pred.shape[0] != 2:
            raise ValueError(f'ConfusionMatrix only supports binary multi-label classification as of now.')

        if self.multilabel and len(pred.shape) > 2 and pred.shape[1] != 2:
            raise ValueError(f'ConfusionMatrix only supports binary multi-label classification as of now.')

        if self.multilabel:
            self.prfa.add(pred, true)
            unique_mapping = ((2 * true + pred) + 4 * _torch.arange(self.num_classes, device=self.device)).flatten()
            matrix = _torch.bincount(unique_mapping, minlength=4 * self.num_classes).reshape(
                self.num_classes,
                2,
                2
            )
        else:
            unique_mapping = (true.view(-1) * self.num_classes + pred.view(-1)).to(_torch.long)
            matrix = _torch.bincount(unique_mapping, minlength=self.num_classes ** 2).reshape(
                self.num_classes,
                self.num_classes
            )
        self.matrix += matrix

    def precision(self, average=True):
        if self.multilabel:
            return self.prfa.precision

        precision = _torch.diag(self.matrix) / _torch.maximum(self.matrix.sum(axis=0), self.matrix_eps)
        if average:
            return (sum(precision) / self.num_classes).item()
        return precision

    def recall(self, average=True):
        if self.multilabel:
            return self.prfa.recall

        recall = _torch.diag(self.matrix) / _torch.maximum(self.matrix.sum(axis=1), self.matrix_eps)
        if average:
            return (sum(recall) / self.num_classes).item()
        return recall

    def f1(self, average=True):
        if self.multilabel:
            return self.prfa.f1
        p = self.precision(average)
        r = self.recall(average)
        if average:
            return 2 * p * r / max(p + r, self.eps)
        return 2 * p * r / torch.maximum(p + r, self.matrix_eps)

    def accuracy(self):
        if self.multilabel:
            return self.prfa.accuracy
        return self.matrix.trace().item() / max(self.matrix.sum().item(), self.eps)

    def get(self):
        if self.multilabel:
            return self.prfa.get()
        return [round(self.accuracy(), self.num_precision), round(self.f1(), self.num_precision),
                round(self.precision(), self.num_precision), round(self.recall(), self.num_precision)]

    def dist_gather(self, device='cpu'):
        if self.multilabel:
            self.prfa.dist_gather(device=device)
        serial = self.matrix.clone().detach().to(device)
        _dist.all_reduce(serial, op=_dist.ReduceOp.SUM)
        self.matrix = serial.cpu()



import json as _json
import multiprocessing as _mp
import os as _os
import traceback as _tb
from typing import Callable

import numpy as _np
import torch as _torch
from torch.utils.data._utils.collate import default_collate as _default_collate
from functools import partial as _partial

from easytorch.utils.logger import success

LOG_FREQ = 100


def _job(total, func, i, f):
    print(f"Working on: [ {i}/{total} ]", end='\n' if i % LOG_FREQ == 0 else '\r')
    try:
        return func(f)
    except Exception as e:
        _tb.print_exc()
        print(f"{f} ### {e}")


def multiRun(nproc: int, data_list: list, func: Callable) -> list:
    _files = []
    for ix, file in enumerate(data_list, 1):
        _files.append([ix, file])

    with _mp.Pool(processes=nproc) as pool:
        return list(
            pool.starmap(_partial(_job, len(_files), func), _files)
        )


def safe_collate(batch):
    r"""Safely select batches/skip dataset_cls(errors in file loading."""
    return _default_collate([b for b in batch if b])


def num_workers(args, loader_args, distributed=False):
    if distributed:
        return (loader_args['num_workers'] + args['num_gpus'] - 1) // args['num_gpus']
    return loader_args.get('num_workers', 0)


def batch_size(args, loader_args, distributed=False):
    if distributed:
        loader_args['batch_size'] = loader_args['batch_size'] // args['num_gpus']
    return loader_args['batch_size']


def seed_worker(worker_id):
    seed = (int(_torch.initial_seed()) + worker_id) % (2 ** 32 - 1)
    _np.random.seed(seed)


def _et_data_job(file, mode, args, dspec, dataset_cls, diskcache):
    dataset = dataset_cls(mode=mode, **args)
    try:
        dataset.add(files=[file], diskcache=diskcache, verbose=False, **dspec)
    except Exception as e:
        _tb.print_exc()
        print(f"{file} ### {e}")
    return dataset


def et_data_job(mode, args, dspec, dataset_cls, total, verbose, diskcache, i, file):
    if verbose:
        print(f"Working on: [ {i} / {total} ]", end='\n' if i % LOG_FREQ == 0 else '\r')
    return _et_data_job(file, mode, args, dspec, dataset_cls, diskcache)


def multi_load(mode, files, dataspec, args, dataset_cls, diskcache=None) -> list:
    r"""Note: Only works with easytorch's default args from easytorch import args"""
    _files = []
    for ix, f in enumerate(files, 1):
        _files.append([ix, f])

    if args.get("multi_load") and len(files) > 1:
        nw = min(num_workers(args, args, args.get('use_ddp')), len(_files))
        with _mp.Pool(processes=max(1, nw)) as pool:
            dataset_list = list(
                pool.starmap(
                    _partial(
                        et_data_job, mode, args, dataspec, dataset_cls, len(_files), args.get('verbose'), diskcache
                    ),
                    _files
                )
            )
            return [d for d in dataset_list if len(d) >= 1]

    dataset_list, total = [], len(_files)
    for ix, file in _files:
        if args.get('verbose'):
            print(f"Loading... {ix}/{total}", end='\n' if ix % LOG_FREQ == 0 else '\r')
        dataset_list.append(_et_data_job(file, mode, args, dataspec, dataset_cls))

    return [d for d in dataset_list if len(d) >= 1]


def pooled_load(split_key, dataspecs, dspec_splits, args, dataset_cls, diskcache=None, load_sparse=False) -> list:
    r"""
    Note: Only works with easytorch's default args from easytorch import args
    This method takes multiple dataspecs and pools the first splits of all the datasets.
    So that we can train one single model on all the datasets. It will automatically refer correct data files,
        no need to move files in single folder.
    """
    all_d = []
    for dspec, split in zip(dataspecs, dspec_splits):
        split = _json.loads(open(dspec['split_dir'] + _os.sep + split).read())
        files = split.get(split_key, [])[:args['load_limit']]

        if load_sparse:
            all_d += multi_load(split_key, files, dspec, args, dataset_cls)
        else:
            if len(all_d) <= 0:
                all_d.append(
                    dataset_cls(mode=split_key, limit=args['load_limit'], **args)
                )
            all_d[0].add(files=files, diskcache=diskcache, verbose=args['verbose'], **dspec)

    success(f'\nPooled {len(all_d)} dataset loaded.', args['verbose'] and len(all_d) > 1)
    return all_d
